from ezdxf.math import Vec3
from scipy.interpolate import interp1d

from navibridge.api import MidasAPI
from navibridge.api.midas_api import MidasDRS
from navibridge.fem_base import Material
from navibridge.fem_base.section import TubeSection, BoxSection, ISection
from navibridge.zoo import TrussUni, SectionGroup
from navibridge.zoo.tower import easy_tower


def h_func_uni(x):
    return 2500


def w_func_uni4(x):
    return 2500


def main(
        wk_name,
        sp_n,
        tower_h,
        k_l,
        k_w,
        truss_l,
        truss_h,
        truss_w,
        slope,
        deltH,
        targetA,
):
    truss_n = int(sum(sp_n))
    func_w_1 = interp1d([0, -1000000], [truss_w, truss_w + 2 * 1000000 * k_w], kind='linear')
    func_l_1 = interp1d([0, -1000000], [truss_l, truss_l + 2 * 1000000 * k_l], kind='linear')
    sps = [n * truss_l * 1e-3 for n in sp_n]
    pi_xx = [0]
    for s in sps[0:-1]:
        pi_xx.append(pi_xx[-1] + s * 1000)
    pi_xx.remove(0)
    pi_zz = [slope * x - (truss_h + deltH) for x in pi_xx]
    tower_n = 1
    m1 = Material(1, 2.06e5, 7.85e-9, 1.2e-5, 0.3, 'Q420')
    m2 = Material(2, 2.06e5, 7.85e-9, 1.2e-5, 0.3, 'Q355')

    md = TrussUni(slope=slope, cell_l=[truss_l, ] * truss_n, cell_h_fuc=h_func_uni, cell_w_fuc=w_func_uni4, pier_loc=pi_xx)
    md.mat_list[1] = m1
    md.mat_list[2] = m2
    for i in range(len(tower_h)):
        x, y = list(zip(pi_xx, pi_zz))[i]
        md += easy_tower(tower_n, Vec3(x, 0, y), tower_h[tower_n - 1], -max(tower_h), func_l_1, func_w_1,
                         delta_l=truss_l, delta_h=deltH, s=slope, targe_ang=targetA)
        tower_n += 1
    ss = SectionGroup()
    ss.add(TubeSection(Id=1, Name="D1", offset=(0, 0), d=400, t=8))
    ss.add(TubeSection(Id=2, Name="D2", offset=(0, 0), d=300, t=6))
    ss.add(TubeSection(Id=3, Name="D2", offset=(0, 0), d=300, t=6))
    ss.add(TubeSection(Id=4, Name="D2", offset=(0, 0), d=200, t=4))
    ss.add(TubeSection(Id=9, Name="D9", offset=(0, 0), d=200, t=4))

    ss.add(ISection(Id=11, Name="上弦杆", offset=(0, 0), w1=200, w2=200, w3=400, t1=13, t2=13, t3=8))
    ss.add(ISection(Id=12, Name="下弦杆", offset=(0, 0), w1=200, w2=200, w3=400, t1=13, t2=13, t3=8))
    ss.add(ISection(Id=21, Name="上平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
    ss.add(ISection(Id=22, Name="下平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
    ss.add(ISection(Id=23, Name="腹杆", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))

    api = MidasAPI(project=wk_name, root_path=".", init=True)
    api.load_config(
        rq=5.25,
        dw=1.6,
        snow=0.3,
        w1=20,
        w2=30,
        temp_up=15,
        temp_down=-10,
        e1=MidasDRS("A", "I0", "VI", 0.45, 0.03, "E1", False, 6, 0.06),
        e2=MidasDRS("A", "I0", "VI", 0.45, 0.03, "E2", False, 6, 0.06),
    )
    api.config(
        truss_h=truss_h,
        truss_l=truss_l,
        slope=slope,
        k_l=k_l,
        k_w=k_w,
    )
    api.run_fem(ss + md)
    api.write_dxf('test.dxf')


if __name__ == '__main__':
    sps = [51.25, ] + [60, ] * 3 + [57.5, ] * 2 + [60, ] * 3 + [57.5] + [1.25]
    spn = [n / 2.5 for n in sps]
    th = [13.5, 25.75, 37.5, 46.5, 42, 37.5, 29.5, 42.0, 42, 33.5]
    main(wk_name="FG2",
         sp_n=spn,
         tower_h=th,
         k_l=4 / 100,
         k_w=5 / 100,
         truss_l=2500,
         truss_h=2500,
         truss_w=2500,
         slope=-0.025,
         deltH=800,
         targetA=30,
         )



from typing import List

import pint
from PyAngle import Angle
from ezdxf.math import Vec3
from scipy.optimize import bisect

from .fem import FEM
from ..fem_base import Node, Element, Fix

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def cal_angle(fw, fl, zup, zlow):
    pts_up = [
        Vec3(+fl(zup) * 0.5, +fw(zup) * 0.5, zup),
        Vec3(-fl(zup) * 0.5, +fw(zup) * 0.5, zup),
        Vec3(+fl(zup) * 0.5, -fw(zup) * 0.5, zup),
    ]
    pts_low = [
        Vec3(+fl(zlow) * 0.5, +fw(zlow) * 0.5, zlow),
        Vec3(-fl(zlow) * 0.5, +fw(zlow) * 0.5, zlow),
        Vec3(+fl(zlow) * 0.5, -fw(zlow) * 0.5, zlow),
    ]
    O = pts_low[0]
    M = pts_up[0]
    A = (M + pts_up[1]) * 0.5
    B = (M + pts_up[2]) * 0.5
    a = Angle.from_rad((M - O).angle_between((A - O))).to_degrees()
    b = Angle.from_rad((M - O).angle_between((B - O))).to_degrees()
    return min(a, b)


def target_func(deltaX, x0, fw, fl, ang):
    return cal_angle(fw, fl, x0, x0 - deltaX) - ang


def create_dist_list(fw, fl, height, ang):
    ret = [0]
    while sum(ret) < height:
        z0 = -sum(ret)
        r = bisect(target_func, 100, 20000, args=(z0 * 1000, fw, fl, ang))
        int_part = int(r / 500)
        ret.append(int_part * 500 * 1e-3)
    ret.remove(ret[0])
    return ret


def easy_tower(tower_id, cc, height, ref_ground, l_fun, w_fun, delta_h, delta_l, s, targe_ang=30.0):
    dist = create_dist_list(w_fun, l_fun, height, targe_ang)
    fem_list = []
    for i, d in enumerate(dist):
        z0 = -sum(dist[0:i]) * 1000
        z1 = z0 - d * 1000
        width0 = w_fun(z0)
        length0 = l_fun(z0)
        width1 = w_fun(z1)
        length1 = l_fun(z1)
        cc0 = cc + Vec3(0, 0, z0)
        if i == (len(dist) - 1):
            add_fix = True
        else:
            add_fix = False
        if i == 0:
            add_top = True
        else:
            add_top = False
        if d >= 15:
            layerN = 3
        elif d >= 7.5:
            layerN = 2
        else:
            layerN = 1
        for px, xi in enumerate([-1, 1]):
            for py, yi in enumerate([-1, 1]):
                bi = f"{px:01b}{py:01b}"
                dec = int(bi, 2) + 1
                if tower_id % 2 == 0:
                    dec += 5
                basen = int(((tower_id - 1) // 2 + 1) * 1e5 + dec * 1e4 + (i + 1) * 1e2)
                md = TriTower(width0, length0, width1, length1, d * 1000, cc0, basen,
                              xi, yi, layerN, ref_ground, [1, 2, 3, 4], slope=s, delt_H=delta_h, delt_L=delta_l,
                              add_fix=add_fix, add_top_brace=add_top)
                fem_list.append(md)
    return sum(fem_list, FEM())


class TriTower(FEM):
    def __init__(self, W0, L0, W1, L1, H, center: Vec3, base_n: int, xdir: int, ydir: int, layer: int,
                 ground_level: float, sect_list, slope,
                 delt_H: float, delt_L: float, add_top_brace=False, add_fix=False):
        """
        生成塔节段
        """
        super().__init__()
        self.H = H
        self.ground_e = ground_level
        self.base_n = base_n
        self.cc = center
        self.xdir = xdir
        self.ydir = ydir
        self.pO = center + Vec3(xdir * 0.5 * L1, ydir * 0.5 * W1, -H)
        self.pA = center + Vec3(xdir * 0.5 * L0, ydir * 0.5 * W0, 0)
        self.pB = center + Vec3(xdir * 0.5 * L0, 0, 0)
        self.pC = center + Vec3(0, ydir * 0.5 * W0, 0)
        self.pD = center + Vec3(xdir * 0.5 * L0, ydir * 0.5 * W0, delt_H - xdir * 0.5 * L0 * (abs(slope)))
        self.pE = center + Vec3(xdir * (0.5 * L0 + delt_L), ydir * 0.5 * W0, delt_H + xdir * (0.5 * L0 + delt_L) * (-(abs(slope))))
        self.OA = (self.pA - self.pO)
        self.OB = (self.pB - self.pO)
        self.OC = (self.pC - self.pO)
        self.AB = (self.pB - self.pA)
        self.AC = (self.pC - self.pA)
        self.layer = layer
        self.e2windx = []
        self.e2windy = []
        self.slist = sect_list
        # self.add_top_brace = add_top_brace
        self.tri_gen(add_fix, add_top_brace)

    def tri_gen(self, is_fix: bool, add_top_brace: bool):
        main_tube_secn = self.slist[0]
        dia_tube_secn = self.slist[1:]
        others = 9
        eid = 0
        killerAB = (self.OA - self.OB).magnitude > 3 * (self.OA.magnitude / self.layer)
        killerAC = (self.OA - self.OC).magnitude > 3 * (self.OA.magnitude / self.layer)
        # OAC
        for n in range(self.layer + 1):
            if n == 0:
                self.add(Node(self.base_n + 1, self.pO))
                if is_fix:
                    self.fix_list[self.base_n + 1] = Fix(self.base_n + 1, {'all': 0})
            else:
                self.add(Node(self.base_n + 1 + n * 10, self.pO + self.OA * (n / self.layer)))
                self.add(Node(self.base_n + 2 + n * 10, self.pO + self.OB * (n / self.layer)))
                self.add(Node(self.base_n + 3 + n * 10, self.pO + self.OC * (n / self.layer)))
                if n == 1:
                    eid = self.quick_elem([11, 1], self.base_n + 1 + n * 10, main_tube_secn, 1)
                    # if self.ydir > 0:
                    #     self.e2windy.append(self.get_e_loc(self.base_n + 1 + n * 10))
                    # if self.xdir > 0:
                    #     self.e2windx.append(self.get_e_loc(self.base_n + 1 + n * 10))
                    eid = self.quick_elem([12, 1], eid, dia_tube_secn[0], 1)
                    eid = self.quick_elem([13, 1], eid, dia_tube_secn[0], 1)
                else:
                    eid = self.quick_elem([1 + n * 10, 1 + (n - 1) * 10], eid, main_tube_secn, 1)
                    # if self.ydir > 0:
                    #     self.e2windy.append(self.get_e_loc(eid - 1))
                    # if self.xdir > 0:
                    #     self.e2windx.append(self.get_e_loc(eid - 1))
                    eid = self.quick_elem([2 + n * 10, 2 + (n - 1) * 10], eid, dia_tube_secn[0], 2)
                    eid = self.quick_elem([3 + n * 10, 3 + (n - 1) * 10], eid, dia_tube_secn[0], 2)
                if (n + 1) / (self.layer + 1) > 2 / 3:
                    if killerAB and killerAC:
                        curA = self.pO + self.OA * (n / self.layer)
                        curB = self.pO + self.OB * (n / self.layer)
                        curC = self.pO + self.OC * (n / self.layer)
                        self.add(Node(self.base_n + 4 + n * 10, curB - 0.5 * self.AB))
                        self.add(Node(self.base_n + 5 + n * 10, curB + 0.5 * (curC - curB)))
                        self.add(Node(self.base_n + 6 + n * 10, curC - 0.5 * self.AC))
                        eid = self.quick_elem(
                            [1 + n * 10, 4 + n * 10, 2 + n * 10, 5 + n * 10, 3 + n * 10, 6 + n * 10, ],
                            eid, others, 2, True)
                        eid = self.quick_elem([4 + n * 10, 5 + n * 10, 6 + n * 10, ], eid, others, 2, True)
                        if self.node_list.keys().__contains__(self.base_n + (n - 1) * 10 + 4):
                            eid = self.quick_elem([2 + (n - 1) * 10, 4 + n * 10, 4 + (n - 1) * 10, 1 + n * 10, ], eid,
                                                  others, 2)
                            eid = self.quick_elem([3 + (n - 1) * 10, 6 + n * 10, 6 + (n - 1) * 10, 1 + n * 10, ], eid,
                                                  others, 2)
                        else:
                            eid = self.quick_elem([2 + (n - 1) * 10, 4 + n * 10, 1 + (n - 1) * 10, ], eid, others, 2)
                            eid = self.quick_elem([3 + (n - 1) * 10, 6 + n * 10, 1 + (n - 1) * 10, ], eid, others, 2)
                    elif killerAB:
                        curA = self.pO + self.OA * (n / self.layer)
                        curB = self.pO + self.OB * (n / self.layer)
                        curC = self.pO + self.OC * (n / self.layer)
                        self.add(Node(self.base_n + 4 + n * 10, curB - 0.5 * self.AB))
                        self.add(Node(self.base_n + 5 + n * 10, curB + 0.5 * (curC - curB)))
                        eid = self.quick_elem([1 + n * 10, 4 + n * 10, 2 + n * 10, 5 + n * 10, 3 + n * 10, ],
                                              eid, others, 2, True)
                        eid = self.quick_elem([4 + n * 10, 5 + n * 10, 1 + n * 10], eid, others, 2)
                        if self.node_list.keys().__contains__(self.base_n + (n - 1) * 10 + 4):
                            eid = self.quick_elem([2 + (n - 1) * 10, 4 + n * 10, 4 + (n - 1) * 10, 1 + n * 10, ], eid,
                                                  others, 2)
                        else:
                            eid = self.quick_elem([2 + (n - 1) * 10, 4 + n * 10, 1 + (n - 1) * 10, ], eid, others, 2)
                        eid = self.quick_elem([1 + n * 10, 3 + (n - 1) * 10], eid, others, 2)
                    elif killerAC:
                        curA = self.pO + self.OA * (n / self.layer)
                        curB = self.pO + self.OB * (n / self.layer)
                        curC = self.pO + self.OC * (n / self.layer)
                        self.add(Node(self.base_n + 5 + n * 10, curB + 0.5 * (curC - curB)))
                        self.add(Node(self.base_n + 6 + n * 10, curC - 0.5 * self.AC))
                        eid = self.quick_elem([1 + n * 10, 2 + n * 10, 5 + n * 10, 3 + n * 10, 6 + n * 10, ],
                                              eid, others, 2, True)
                        eid = self.quick_elem([1 + n * 10, 5 + n * 10, 6 + n * 10, ], eid, others, 2)
                        if self.node_list.keys().__contains__(self.base_n + (n - 1) * 10 + 6):
                            eid = self.quick_elem([3 + (n - 1) * 10, 6 + n * 10, 6 + (n - 1) * 10, 1 + n * 10, ], eid,
                                                  others, 2)
                        else:
                            eid = self.quick_elem([3 + (n - 1) * 10, 6 + n * 10, 1 + (n - 1) * 10, ], eid, others, 2)
                        eid = self.quick_elem([1 + n * 10, 2 + (n - 1) * 10], eid, others, 2)
                    else:
                        if n == self.layer:
                            eid = self.quick_elem([2 + n * 10, 1 + n * 10, 3 + n * 10], eid, dia_tube_secn[1], 2)  # 顶层三角
                            eid = self.quick_elem([2 + n * 10, 3 + n * 10], eid, dia_tube_secn[2], 2)
                        else:
                            eid = self.quick_elem([1 + n * 10, 2 + n * 10, 3 + n * 10], eid, others, 2, True)
                        if self.layer != 1:
                            eid = self.quick_elem([1 + n * 10, 2 + (n - 1) * 10], eid, others, 2)
                            eid = self.quick_elem([1 + n * 10, 3 + (n - 1) * 10], eid, others, 2)
                else:
                    eid = self.quick_elem([1 + n * 10, 2 + n * 10, 3 + n * 10], eid, others, 2, True)
                    if n != 1:
                        eid = self.quick_elem([1 + n * 10, 2 + (n - 1) * 10], eid, others, 2)
                        eid = self.quick_elem([1 + n * 10, 3 + (n - 1) * 10], eid, others, 2)
        if add_top_brace:  # 首层加支撑
            self.add(Node(self.base_n + 51, self.pD))
            self.add(Node(self.base_n + 52, self.pE))
            self.add(Element(e=self.base_n + 1 + 50, secn=main_tube_secn, mat=1,
                             nodes=[
                                 self.node_list[self.base_n + 11],
                                 self.node_list[self.base_n + 51]]
                             ))
            self.add(Element(e=self.base_n + 1 + 51, secn=2, mat=1,
                             nodes=[
                                 self.node_list[self.base_n + 1],
                                 self.node_list[self.base_n + 52]
                             ]))

    def quick_elem(self, nlist, e0, secn, matid, closed=False):
        for i, n in enumerate(nlist):
            if n != nlist[0]:
                if type(secn) is not List:
                    s = secn
                else:
                    s = secn[i - 1]
                self.add(Element(e0, secn=s, mat=matid, nodes=[
                    self.node_list[nlist[i - 1] + self.base_n],
                    self.node_list[n + self.base_n]]
                                 ))
                e0 += 1
        if closed:
            if type(secn) is not List:
                s = secn
            else:
                s = secn[-1]
            self.add(Element(e0, secn=s, mat=matid, nodes=[
                self.node_list[nlist[- 1] + self.base_n],
                self.node_list[nlist[0] + self.base_n]]
                             ))
            e0 += 1
        return e0

    def get_e_loc(self, eid):
        n1 = self.elem_list[eid].nlist[0]
        n2 = self.elem_list[eid].nlist[1]
        return eid, 0.5 * (n1.vec + n2.vec)

    @staticmethod
    def getFg(gv, Ud, C, D, rho=Q_(1.25, ureg.kg / ureg.m ** 3)):
        """
        静风荷载
        :param gv:
        :param Ud:
        :param C:
        :param D:
        :param rho:
        :return:
        """
        Ug = Q_(gv * Ud, ureg.m / ureg.s)
        F = (0.5 * rho * Ug ** 2 * C * Q_(D, ureg.m)).to(ureg.N / ureg.m)
        return F

    @staticmethod
    def GetUd(U10, Z, kc, kf, alpha0):
        """
        基本风速换算基准风速
        :param U10: 基本风速
        :param Z: 基准高度
        :param kc: 地表类别换算系数，D=0.564
        :param kf: 抗风风险系数 R2=1.02
        :param alpha0: 地表粗糙度系数
        :return:
        """
        Us10 = kc * U10 * ureg.m / ureg.s
        Ud = kf * (Z / 10) ** alpha0 * Us10
        return Ud

    def add(self, obj):
        if isinstance(obj, Node):
            collect = self.node_list
        elif isinstance(obj, Element):
            if self.node_list.keys().__contains__(obj.nlist[0].id) and self.node_list.keys().__contains__(obj.nlist[1].id):
                collect = self.elem_list
                if obj.secn in [1, ] and int(obj.id % 1e5 // 1e4) in [2, 4, 7, 9]:
                    self.left_elem_list[obj.id] = obj
                if obj.secn in [1, ] and int(obj.id % 1e5 // 1e4) in [1, 2, 6, 7]:
                    self.front_elem_list[obj.id] = obj
            else:
                raise ValueError("节点不存在.")
        else:
            raise TypeError
        collect[obj.id] = obj



import pint

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def GetUd(U10, Z, kc, kf, alpha0):
    """
    基本风速换算基准风速
    :param U10: 基本风速
    :param Z: 基准高度
    :param kc: 地表类别换算系数，D=0.564
    :param kf: 抗风风险系数 R2=1.02
    :param alpha0: 地表粗糙度系数
    :return:
    """
    Us10 = kc * U10 * ureg.m / ureg.s
    Ud = kf * (Z / 10) ** alpha0 * Us10
    return Ud


def getFg(gv, Ud, C, D, rho=Q_(1.25, ureg.kg / ureg.m ** 3)):
    """
    静风荷载
    :param gv:
    :param Ud:
    :param C:
    :param D:
    :param rho:
    :return:
    """
    Ug = Q_(gv * Ud, ureg.m / ureg.s)
    F = (0.5 * rho * Ug ** 2 * C * Q_(D, ureg.m)).to(ureg.N / ureg.m)
    return F



import numpy as np

def PrintSPD(Ts, S, file):
    with open(file, "w+") as fid:
        fid.write("*UNIT,GRAV\n")
        fid.write("*TYPE,ACCEL\n")
        fid.write("*Data\n")
        for i, t in enumerate(Ts):
            fid.write("%.4E,%.6f\n" % (t, S[i]))


def GetS(Smax, Tg, T):
    T0 = 0.1
    if T < T0:
        return Smax * (0.6 * T / T0 + 0.4)
    elif T < Tg:
        return Smax
    else:
        return Smax * (Tg / T)


def GetDES(Smax, Tg, Ta=0.0, Tb=10.0, Tstep=0.01):
    Tlist = np.linspace(Ta, Tb, int(np.round((Tb - Ta) / Tstep + 1)))
    Slist = [GetS(Smax, Tg, ti) for ti in Tlist]
    return [Tlist.tolist(), Slist]


def GetSmax(Ci, Cs, Cd, A):
    return 2.5 * Ci * Cs * Cd * A


def GetCd(xi):
    return max(1 + (0.05 - xi) / (0.08 + 1.6 * xi), 0.55)


if __name__ == "__main__":
    Cd0 = GetCd(0.03)
    print("Cd=%.4f" % Cd0)
    Smax_E1x = GetSmax(0.34, 1.0, Cd0, 0.2)
    Smax_E1z = GetSmax(0.34, 0.6, Cd0, 0.2)
    Smax_E2x = GetSmax(1.3, 1.0, Cd0, 0.2)
    Smax_E2z = GetSmax(1.3, 0.6, Cd0, 0.2)
    print("Smax_E1x = %.4f" % Smax_E1x)
    print("Smax_E1z = %.4f" % Smax_E1z)
    print("Smax_E2x = %.4f" % Smax_E2x)
    print("Smax_E2z = %.4f" % Smax_E2z)
    DES_E1x = GetDES(Smax_E1x, 0.45)
    DES_E1z = GetDES(Smax_E1z, 0.45)
    DES_E2x = GetDES(Smax_E2x, 0.45)
    DES_E2z = GetDES(Smax_E2z, 0.45)

    PrintSPD(DES_E1x[0], DES_E1x[1], "../../bin/out/E1水平设计反应谱.spd")
    PrintSPD(DES_E2x[0], DES_E2x[1], "../../bin/out/E2水平设计反应谱.spd")



from typing import List
from ezdxf.math import Vec3
from scipy.interpolate import interp1d
from .ansys_api import AnsysAPI
from ..fem_base import Material
from ..zoo import SectionGroup, TrussUni
from ..fem_base.section import TubeSection, BoxSection, ISection
from ..zoo.tower import easy_tower


class AnsysTask:
    def __init__(self, root_path, wk_name, sp_n: List[float], tower_h: List[float], k_l, k_w,
                 truss_l, truss_h, truss_w, slope, tower_sec_category, truss_sec_category, deltH=800):
        super().__init__()
        self.config = {}
        self.loads_config = {}
        self.dimension_config = {}
        self.root_path = root_path
        self.wk_name = wk_name
        self.sp_n = sp_n
        self.tower_h = tower_h
        self.k_l = k_l
        self.k_w = k_w
        self.truss_l = truss_l
        self.truss_h = truss_h
        self.truss_w = truss_w
        self.slope = slope
        self.deltH = deltH
        self.dimension_config['length'] = sum(sp_n) * truss_l * 1e-3
        self.dimension_config['truss'] = (truss_h, truss_w, truss_l)
        self.dimension_config['slope'] = slope
        self.dimension_config['kl_kw'] = (k_l, k_w)
        self.tower_sect_category = tower_sec_category
        self.truss_sec_category = truss_sec_category

    def h_func_uni(self, x):
        return self.truss_h

    def w_func_uni4(self, x):
        return self.truss_w

    def perform_task(self):
        sp_n = self.sp_n
        tower_h: List[float] = self.tower_h
        k_l, k_w = self.k_l, self.k_w
        truss_l, truss_h, truss_w = self.truss_l, self.truss_h, self.truss_w
        slope = self.slope
        deltH = self.deltH
        truss_n = int(sum(sp_n))
        func_w_1 = interp1d([0, -1000000], [truss_w, truss_w + 2 * 1000000 * k_w], kind='linear')
        func_l_1 = interp1d([0, -1000000], [truss_l, truss_l + 2 * 1000000 * k_l], kind='linear')
        sps = [n * truss_l * 1e-3 for n in sp_n]
        pi_xx = [0]
        for s in sps[0:-1]:
            pi_xx.append(pi_xx[-1] + s * 1000)
        pi_xx.remove(0)
        pi_zz = [slope * x - (truss_h + deltH) for x in pi_xx]
        tower_n = 1
        md = TrussUni(slope=slope, cell_l=[truss_l, ] * truss_n,
                      cell_h_fuc=self.h_func_uni,
                      cell_w_fuc=self.w_func_uni4,
                      pier_loc=pi_xx
                      )
        for i in range(len(tower_h)):
            x, y = list(zip(pi_xx, pi_zz))[i]
            md += easy_tower(tower_n, Vec3(x, 0, y), tower_h[tower_n - 1], -max(tower_h), func_l_1, func_w_1,
                             delta_l=truss_l, delta_h=deltH, s=slope)
            tower_n += 1
        ss = SectionGroup()
        if self.tower_sect_category == 'ClassA':
            ss.add(TubeSection(Id=1, Name="D1", offset=(0, 0), d=400, t=10))
            ss.add(TubeSection(Id=2, Name="D2", offset=(0, 0), d=300, t=6))
            ss.add(TubeSection(Id=3, Name="D3", offset=(0, 0), d=300, t=6))
            ss.add(TubeSection(Id=4, Name="D4", offset=(0, 0), d=200, t=4))
            ss.add(TubeSection(Id=9, Name="D9", offset=(0, 0), d=200, t=4))
        elif self.tower_sect_category == 'ClassB':
            ss.add(TubeSection(Id=1, Name="D1", offset=(0, 0), d=600, t=12))
            ss.add(TubeSection(Id=2, Name="D2", offset=(0, 0), d=400, t=10))
            ss.add(TubeSection(Id=3, Name="D3", offset=(0, 0), d=400, t=10))
            ss.add(TubeSection(Id=4, Name="D4", offset=(0, 0), d=300, t=6))
            ss.add(TubeSection(Id=9, Name="D9", offset=(0, 0), d=200, t=4))
        elif self.tower_sect_category == 'ClassC':
            ss.add(TubeSection(Id=1, Name="D1", offset=(0, 0), d=800, t=16))
            ss.add(TubeSection(Id=2, Name="D2", offset=(0, 0), d=600, t=12))
            ss.add(TubeSection(Id=3, Name="D3", offset=(0, 0), d=600, t=12))
            ss.add(TubeSection(Id=4, Name="D4", offset=(0, 0), d=300, t=6))
            ss.add(TubeSection(Id=9, Name="D9", offset=(0, 0), d=200, t=4))
        else:
            raise NotImplementedError

        if self.truss_sec_category == 'ClassA':
            ss.add(ISection(Id=11, Name="上弦杆", offset=(0, 0), w1=200, w2=200, w3=400, t1=13, t2=13, t3=8))
            ss.add(ISection(Id=12, Name="下弦杆", offset=(0, 0), w1=200, w2=200, w3=400, t1=13, t2=13, t3=8))
            ss.add(ISection(Id=21, Name="上平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
            ss.add(ISection(Id=22, Name="下平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
            ss.add(ISection(Id=23, Name="腹杆", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
        elif self.truss_sec_category == 'ClassB':
            ss.add(BoxSection(Id=11, Name="上弦杆", offset=(0, 0), w1=400, w2=250, t1=12, t2=12, t3=12, t4=12))
            ss.add(BoxSection(Id=12, Name="下弦杆", offset=(0, 0), w1=400, w2=250, t1=12, t2=12, t3=12, t4=12))
            ss.add(ISection(Id=21, Name="上平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
            ss.add(ISection(Id=22, Name="下平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
            ss.add(ISection(Id=23, Name="腹杆", offset=(0, 0), w1=175, w2=175, w3=250, t1=11, t2=11, t3=7))
        elif self.truss_sec_category == 'ClassC':
            ss.add(BoxSection(Id=11, Name="上弦杆", offset=(0, 0), w1=400, w2=300, t1=18, t2=18, t3=18, t4=18))
            ss.add(BoxSection(Id=12, Name="下弦杆", offset=(0, 0), w1=400, w2=300, t1=18, t2=18, t3=18, t4=18))
            ss.add(ISection(Id=21, Name="上平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
            ss.add(ISection(Id=22, Name="下平联", offset=(0, 0), w1=150, w2=150, w3=200, t1=9, t2=9, t3=6))
            ss.add(ISection(Id=23, Name="腹杆", offset=(0, 0), w1=200, w2=200, w3=300, t1=14, t2=14, t3=9))
        else:
            raise NotImplementedError
        ss.add(Material(1, 2.06e5, 7.85e-9 * self.loads_config['dc_factor'], 1.2e-5, 0.3, 'Q420'))
        ss.add(Material(2, 2.06e5, 7.85e-9 * self.loads_config['dc_factor'], 1.2e-5, 0.3, 'Q355'))

        api = AnsysAPI(project=self.wk_name, root_path=self.root_path, init=self.config['init'])
        api.config(
            image=self.config['image'],
            only_cmd=self.config['only_cmd'],
            ANSYS_PATH=self.config['ANSYS_PATH'],
        )
        api.run_fem(ss + md)
        api.run_deadload(lcn=1)
        api.run_dw(lcn=2)
        api.run_rq(self.loads_config['RQ'], lcn=6)
        api.run_wind('W1', self.loads_config['W1'], pi_xx, [2. / 3. * h for h in tower_h], max(tower_h), truss_h * 1e-3, lcn=31)
        api.run_wind('W2', self.loads_config['W2'], pi_xx, [2. / 3. * h for h in tower_h], max(tower_h), truss_h * 1e-3, lcn=32)
        api.run_wind('W3', self.loads_config['W3'], pi_xx, [2. / 3. * h for h in tower_h], max(tower_h), truss_h * 1e-3, lcn=33)
        api.run_wind('W4', self.loads_config['W4'], pi_xx, [2. / 3. * h for h in tower_h], max(tower_h), truss_h * 1e-3, lcn=34)
        api.run_wind('W5', self.loads_config['W5'], pi_xx, [2. / 3. * h for h in tower_h], max(tower_h), truss_h * 1e-3, lcn=35)
        api.run_wind('W6', self.loads_config['W6'], pi_xx, [2. / 3. * h for h in tower_h], max(tower_h), truss_h * 1e-3, lcn=36)
        api.run_tu("TU",
                   temp_up=self.loads_config['TempUp'],
                   temp_down=self.loads_config['TempDown'], lcn=40)
        api.run_se('SE', val=self.loads_config['SE'], num_tower=len(tower_h), lcn=5)
        api.run_lcomb("ULS", conf={
            81: [(1, 1.2), (2, 1.2), (5, 1.0), (31, 1.4), (41, 1.05), ],
            82: [(1, 1.2), (2, 1.2), (5, 1.0), (32, 1.4), (41, 1.05), ],
            83: [(1, 1.2), (2, 1.2), (5, 1.0), (33, 1.4), (41, 1.05), ],
            84: [(1, 1.2), (2, 1.2), (5, 1.0), (34, 1.4), (41, 1.05), ],
            85: [(1, 1.2), (2, 1.2), (5, 1.0), (35, 1.4), (41, 1.05), ],
            86: [(1, 1.2), (2, 1.2), (5, 1.0), (36, 1.4), (41, 1.05), ],
        })
        api.init_records(self.dimension_config, self.loads_config)
        api.run_mode(self.loads_config['MODEL'])
        api.run_rs("e1dx", lcn=51, direction="X", Ci=1.3, Cs=1.0, A=self.loads_config['E1'], Tg=0.35)  # 0.05g
        api.run_rs("e1dy", lcn=52, direction="Y", Ci=1.3, Cs=0.6, A=self.loads_config['E1'], Tg=0.25)  # 0.05g
        api.run_rs("e1dz", lcn=53, direction="Z", Ci=1.3, Cs=1.0, A=self.loads_config['E1'], Tg=0.35)  # 0.05g
        api.run_rs("e2dx", lcn=54, direction="X", Ci=1.3, Cs=1.0, A=self.loads_config['E2'], Tg=0.35)  # 0.15g
        api.run_rs("e2dy", lcn=55, direction="Y", Ci=1.3, Cs=0.6, A=self.loads_config['E2'], Tg=0.25)  # 0.15g
        api.run_rs("e2dz", lcn=56, direction="Z", Ci=1.3, Cs=1.0, A=self.loads_config['E2'], Tg=0.35)  # 0.15g
        api.run_rs("e3dx", lcn=57, direction="X", Ci=1.3, Cs=1.0, A=self.loads_config['E3'], Tg=0.35)  # 0.20g
        api.run_rs("e3dy", lcn=58, direction="Y", Ci=1.3, Cs=0.6, A=self.loads_config['E3'], Tg=0.25)  # 0.20g
        api.run_rs("e3dz", lcn=59, direction="Z", Ci=1.3, Cs=1.0, A=self.loads_config['E3'], Tg=0.35)  # 0.20g
        api.run_rs("e4dx", lcn=61, direction="X", Ci=1.3, Cs=1.0, A=self.loads_config['E4'], Tg=0.35)  # 0.30g
        api.run_rs("e4dy", lcn=62, direction="Y", Ci=1.3, Cs=0.7, A=self.loads_config['E4'], Tg=0.25)  # 0.30g
        api.run_rs("e4dz", lcn=63, direction="Z", Ci=1.3, Cs=1.0, A=self.loads_config['E4'], Tg=0.35)  # 0.30g
        api.run_rs("e5dx", lcn=64, direction="X", Ci=1.3, Cs=1.0, A=self.loads_config['E5'], Tg=0.35)  # 0.40g
        api.run_rs("e5dy", lcn=65, direction="Y", Ci=1.3, Cs=0.7, A=self.loads_config['E5'], Tg=0.25)  # 0.40g
        api.run_rs("e5dz", lcn=66, direction="Z", Ci=1.3, Cs=1.0, A=self.loads_config['E5'], Tg=0.35)  # 0.40g
        api.run_rs("e6dx", lcn=67, direction="X", Ci=1.3, Cs=1.0, A=self.loads_config['E6'], Tg=0.35)  # 0.40g
        api.run_rs("e6dy", lcn=68, direction="Y", Ci=1.3, Cs=0.7, A=self.loads_config['E6'], Tg=0.25)  # 0.40g
        api.run_rs("e6dz", lcn=69, direction="Z", Ci=1.3, Cs=1.0, A=self.loads_config['E6'], Tg=0.35)  # 0.40g
        api.srss_rs('e1', lcn=11, lc_x=51, lc_y=52, lc_z=53)
        api.srss_rs('e2', lcn=12, lc_x=54, lc_y=55, lc_z=56)
        api.srss_rs('e3', lcn=13, lc_x=57, lc_y=58, lc_z=59)
        api.srss_rs('e4', lcn=14, lc_x=61, lc_y=62, lc_z=63)
        api.srss_rs('e5', lcn=15, lc_x=64, lc_y=65, lc_z=66)
        api.srss_rs('e6', lcn=16, lc_x=67, lc_y=68, lc_z=69)
        api.run_lcomb("ELS", conf={
            91: [(1, 1.0), (2, 1.0), (11, 1.0), ],
            92: [(1, 1.0), (2, 1.0), (12, 1.0), ],
            93: [(1, 1.0), (2, 1.0), (13, 1.0), ],
            94: [(1, 1.0), (2, 1.0), (14, 1.0), ],
            95: [(1, 1.0), (2, 1.0), (15, 1.0), ],
            96: [(1, 1.0), (2, 1.0), (16, 1.0), ],
        })
        config = {
            1: [6, 'Y', "RQuy"],
            2: [31, 'Z', "W1uz"],
            3: [32, 'Z', "W2uz"],
            4: [33, 'Z', "W3uz"],
            5: [34, 'Z', "W4uz"],
            6: [35, 'Z', "W5uz"],
            7: [36, 'Z', "W6uz"],
        }
        api.get_node_deformation(config)
        api.get_elem_forces(1, 81)
        api.get_elem_forces(1, 82)
        api.get_elem_forces(1, 83)
        api.get_elem_forces(1, 84)
        api.get_elem_forces(1, 85)
        api.get_elem_forces(1, 86)
        api.get_elem_forces(1, 91)
        api.get_elem_forces(1, 92)
        api.get_elem_forces(1, 93)
        api.get_elem_forces(1, 94)
        api.get_elem_forces(1, 95)
        api.get_elem_forces(1, 96)
        api.get_elem_forces(11, 81)
        api.get_elem_forces(11, 82)
        api.get_elem_forces(11, 83)
        api.get_elem_forces(11, 84)
        api.get_elem_forces(11, 85)
        api.get_elem_forces(11, 86)
        api.get_elem_forces(11, 91)
        api.get_elem_forces(11, 92)
        api.get_elem_forces(11, 93)
        api.get_elem_forces(11, 94)
        api.get_elem_forces(11, 95)
        api.get_elem_forces(11, 96)
        if self.config['clear']:
            api.clear()

    def configuration(self, **kwargs):
        self.config = {**kwargs}

    def loads(self, **kwargs):
        self.loads_config = {**kwargs}



import os
from abc import ABCMeta, abstractmethod

import pint

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


class BaseAPI(metaclass=ABCMeta):
    def __init__(self, project, root_path, init=True):
        # self.cwd = ""
        self.options = {}
        self.fem: 'FEM' = None
        self.project_name = project
        wdr = os.path.join(root_path, self.project_name)
        self.cwd = os.path.abspath(wdr)
        self.res_dir = os.path.join(os.path.abspath(root_path), f'{project}_result.txt')
        self.init = init
        self.loads = {}

    def config(self, **kwargs):
        self.options = {**kwargs}

    def load_config(self, **kwargs):
        self.loads = {**kwargs}

    def write_dxf(self, dxf_file):
        import ezdxf
        from ezdxf.lldxf.const import DXFTableEntryError
        # 创建一个新的 DXF 文档
        doc = ezdxf.new(dxfversion='R2010')
        for s in self.fem.sect_list.keys():
            sec = self.fem.sect_list[s]
            try:
                doc.layers.add(name=f"{sec.name}", color=sec.id)
            except DXFTableEntryError:
                continue
        # 获取模型空间（用于绘制图形）
        msp = doc.modelspace()
        for ek in self.fem.elem_list.keys():
            ele = self.fem.elem_list[ek]
            # 绘制一条空间直线，指定起点和终点，并设置图层为 "Layer1"
            start_point = [v * 1e-3 for v in ele.nlist[0].vec.xyz]  # 起点坐标
            end_point = [v * 1e-3 for v in ele.nlist[1].vec.xyz]  # 终点坐标
            secn = self.fem.sect_list[ele.secn].name
            msp.add_line(start_point, end_point, dxfattribs={'layer': secn})

        # 保存为一个新的 DXF 文件
        doc.saveas(os.path.join(self.cwd, dxf_file))

    @abstractmethod
    def clear(self):
        pass

    @staticmethod
    def GetUd(U10, Z, kc, kf, alpha0):
        """
        基本风速换算基准风速
        :param U10: 基本风速
        :param Z: 基准高度
        :param kc: 地表类别换算系数，D=0.564
        :param kf: 抗风风险系数 R2=1.02
        :param alpha0: 地表粗糙度系数
        :return:
        """
        Us10 = kc * U10 * ureg.m / ureg.s
        Ud = kf * (Z / 10) ** alpha0 * Us10
        return Ud

    @staticmethod
    def getFg(gv, Ud, C, D, rho=Q_(1.25, ureg.kg / ureg.m ** 3)):
        """
        静风荷载
        :param gv:
        :param Ud:
        :param C:
        :param D:
        :param rho:
        :return:
        """
        Ug = Q_(gv * Ud, ureg.m / ureg.s)
        F = (0.5 * rho * Ug ** 2 * C * Q_(D, ureg.m)).to(ureg.N / ureg.m)
        return F



from typing import Dict

from navibridge.fem_base import Node, Element, Section, Material, Fix


class FEM:
    node_list: Dict[int, 'Node']
    elem_list: Dict[int, 'Element']
    top_elem_list: Dict[int, 'Element']
    left_elem_list: Dict[int, 'Element']
    front_elem_list: Dict[int, 'Element']
    sect_list: Dict[int, 'Section']
    mat_list: Dict[int, 'Material']
    fix_list: Dict[int, 'Fix']

    def __init__(self):
        self._initialize()

    def _initialize(self):
        """
        初始化模型。
        :return:
        """
        self.node_list: {int: Node} = {}
        self.elem_list: {int: Element} = {}
        self.top_elem_list: {int: Element} = {}
        self.left_elem_list: {int: Element} = {}
        self.front_elem_list: {int: Element} = {}
        self.sect_list: {int: Section} = {}
        self.mat_list: {int: Material} = {}
        self.fix_list: {int: Fix} = {}
        self.is_fem = False
        self.apdl = ""

    def __add__(self, other):
        if not isinstance(other, FEM):
            return NotImplemented

        # 检查重复键
        common_keys_node = self.node_list.keys() & other.node_list.keys()
        common_keys_elem = self.elem_list.keys() & other.elem_list.keys()
        common_keys_elem_t = self.top_elem_list.keys() & other.top_elem_list.keys()
        common_keys_elem_l = self.left_elem_list.keys() & other.left_elem_list.keys()
        common_keys_elem_f = self.front_elem_list.keys() & other.front_elem_list.keys()
        common_keys_sect = self.sect_list.keys() & other.sect_list.keys()
        common_keys_mat = self.mat_list.keys() & other.mat_list.keys()
        common_keys_bd = self.fix_list.keys() & other.fix_list.keys()

        if (
                common_keys_node
                or common_keys_elem
                or common_keys_elem_t
                or common_keys_elem_l
                or common_keys_elem_f
                or common_keys_sect
                or common_keys_mat
                or common_keys_bd
        ):
            raise ValueError("Cannot add FEM objects: duplicate keys found.")

        new_fem = FEM()
        new_fem.node_list = {**self.node_list, **other.node_list}
        new_fem.elem_list = {**self.elem_list, **other.elem_list}
        new_fem.top_elem_list = {**self.top_elem_list, **other.top_elem_list}
        new_fem.left_elem_list = {**self.left_elem_list, **other.left_elem_list}
        new_fem.front_elem_list = {**self.front_elem_list, **other.front_elem_list}
        new_fem.sect_list = {**self.sect_list, **other.sect_list}
        new_fem.mat_list = {**self.mat_list, **other.mat_list}
        new_fem.fix_list = {**self.fix_list, **other.fix_list}
        return new_fem

#     def __sub__(self, other):
#         if not isinstance(other, FEM):
#             return NotImplemented
#
#         new_fem = FEM()
#         new_fem.node_list = {k: v for k, v in self.node_list.items() if k not in other.node_list}
#         new_fem.elem_list = {k: v for k, v in self.elem_list.items() if k not in other.elem_list}
#         new_fem.sect_list = {k: v for k, v in self.sect_list.items() if k not in other.sect_list}
#         new_fem.mat_list = {k: v for k, v in self.mat_list.items() if k not in other.mat_list}
#         return new_fem
#



from abc import ABCMeta
from typing import Tuple

from .base import ApdlWriteable


class Section:
    __slots__ = ('id', 'name', '_offset')
    __metaclass__ = ABCMeta

    def __init__(self, Id: int, Name: str, offset: Tuple[float, float]):
        self.id = Id
        self.name = Name
        self._offset = offset

    def __cmp__(self, other: "Section"):
        """compare two Section
        negative if self <  other;
        zero     if self == other;
        positive if self >  other;
        """
        return self.id - other.id

    def set_offset(self, x=0, y=0):
        self._offset = (x, y)

    @property
    def offset(self):
        return self._offset


class ShellSection(Section, ApdlWriteable):
    """
    Shell截面
    """

    __slots__ = ('thickness',)

    thickness: float

    def __init__(self, Id: int, Name: str, offset: Tuple[float, float], th: float):
        super().__init__(Id, Name, offset)
        self.thickness = th

    @property
    def apdl_str(self):
        cmd_str = ''' 
! Shell截面
sect,%i,shell,,%s
secdata, %.5f,1,0.0,3  
secoffset,user,%.5f  
seccontrol,,,, , , ,''' % (self.id, self.name, self.thickness, self._offset[1])
        return cmd_str

    def __str__(self):
        st = "%i : %s : (" % (self.id, self.name)
        for key in self.__slots__:
            st += "%.3f," % getattr(self, key)
        st += ")"
        return st


class ISection(Section, ApdlWriteable):
    """
    I型梁截面
    """

    __slots__ = ('w1', 'w2', 'w3', 't1', 't2', 't3')

    def __init__(self, Id: int, Name: str, offset: Tuple[float, float], **kwargs):
        super().__init__(Id, Name, offset)
        for key, value in kwargs.items():
            if key in self.__slots__:
                setattr(self, key, value)

    def __str__(self):
        st = "%i : %s : (" % (self.id, self.name)
        for key in self.__slots__:
            st += "%.3f," % getattr(self, key)
        st += ")"
        return st

    @property
    def mct_str(self):
        cmd = (f" {self.id}, DBUSER, {self.name}   , CC, 0, 0, 0, 0, 0, 0, YES, NO,"
               + f"H  , 2, {self.w3}, {self.w2}, {self.t3}, {self.t2}, {self.w1}, {self.t1}, 0, 0, 0, 0\n")
        return cmd

    @property
    def apdl_str(self):
        if self._offset[0] == 0 and self._offset[1] == 0:
            cmd_str = '''
! I型梁截面
sect,  %i, BEAM, I, %s, 0   
SECOFFSET, CENT 
secdata,%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,0,0,0,0,0,0''' % (
                self.id, self.name,
                self.w1, self.w2, self.w3, self.t1, self.t2, self.t3)
        else:
            cmd_str = '''
! I型梁截面
sect,  %i, BEAM, I, %s, 0   
secoffset, user, %f, %f 
secdata,%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,0,0,0,0,0,0''' % (
                self.id, self.name, self._offset[0], self._offset[1],
                self.w1, self.w2, self.w3, self.t1, self.t2, self.t3)
        return cmd_str

    @property
    def yg(self):
        h = self.w3 - self.t1 - self.t2
        a1 = self.t1 * self.w1
        a3 = self.t3 * h
        a2 = self.t2 * self.w2
        y1 = 0.5 * self.t1
        y3 = self.t1 + 0.5 * h
        y2 = self.w3 - self.t2 * 0.5
        y0 = (y1 * a1 + y2 * a2 + y3 * a3) / self.area
        return y0

    @property
    def area(self):
        return self.w1 * self.t1 + self.w2 * self.t2 + (self.w3 - self.t1 - self.t2) * self.t3

    @property
    def inertia(self):
        h = self.w3 - self.t1 - self.t2
        a1 = self.t1 * self.w1
        a3 = self.t3 * h
        a2 = self.t2 * self.w2
        y1 = 0.5 * self.t1
        y3 = self.t1 + 0.5 * h
        y2 = self.w3 - self.t2 * 0.5
        y0 = (y1 * a1 + y2 * a2 + y3 * a3) / self.area
        i1 = self.t1 ** 3 * self.w1 / 12 + a1 * (y0 - y1) ** 2
        i2 = self.t2 ** 3 * self.w2 / 12 + a2 * (y0 - y2) ** 2
        i3 = self.t3 * h ** 3 / 12 + a3 * (y0 - y3) ** 2
        i = i1 + i2 + i3
        return i


class BoxSection(Section):
    __slots__ = ('w1', 'w2', 't1', 't2', 't3', 't4')

    def __init__(self, Id: int, Name: str, offset: Tuple[float, float], w1: float, w2: float, t1: float, t2: float, t3: float, t4: float):
        super().__init__(Id, Name, offset)
        self.w1 = w1
        self.w2 = w2
        self.t1 = t1
        self.t2 = t2
        self.t3 = t3
        self.t4 = t4

    @property
    def apdl_str(self):
        cmd_str = f'''
sectype, {self.id}, BEAM, HREC, {self.name}, 0
SECOFFSET, cent
secdata, {self.w1}, {self.w2}, {self.t1}, {self.t2}, {self.t3}, {self.t4}, 0, 0, 0, 0, 0, 0
'''
        return cmd_str

    @property
    def mct_str(self):
        cmd = f" {self.id}, DBUSER    , {self.name}  , CC, 0, 0, 0, 0, 0, 0, YES, NO, B  ,2, {self.w2}, {self.w1}, {self.t3}, {self.t1}, 0, 0, 0, 0, 0, 0\n"
        return cmd


class TubeSection(Section):
    __slots__ = ('diameter', 'thickness')

    def __init__(self, Id: int, Name: str, offset: Tuple[float, float], d: float, t: float):
        super().__init__(Id, Name, offset)
        self.diameter = d
        self.thickness = t

    @property
    def apdl_str(self):
        ro = 0.5 * self.diameter
        ri = ro - self.thickness
        cmd_str = f'''
sectype, {self.id}, BEAM, CTUBE, {self.name}, 0
SECOFFSET, cent
secdata, {ri}, {ro}, 12,0,0,0,0,0,0,0,0,0
'''
        return cmd_str

    @property
    def mct_str(self):
        cmd = f' {self.id}, DBUSER , {self.name} , CC, 0, 0, 0, 0, 0, 0, YES, NO, P  , 2, {self.diameter}, {self.thickness}, 0, 0, 0, 0, 0, 0, 0, 0\n'
        return cmd



from typing import Tuple, List

import numpy as np

from . import Node
from .base import ApdlWriteable


class Element(ApdlWriteable):
    __slots__ = ('e', 'nlist', 'mat', 'etype', 'real', 'secn', '_npts')

    nlist: List[Node]

    def __init__(self, e: int, mat: int, secn: int, nodes: List[Node], etype: int = 188, real: int = 188):
        self.id = e
        self.mat = mat
        self.etype = etype
        self.real = real
        self.secn = secn
        self._npts = len(nodes)
        if self.etype == 188:
            if len(nodes) != 2:
                raise Exception("BEAM188应使用2个Node.")
            else:
                self.nlist = nodes
        elif etype == 181 or etype == 182:
            if len(nodes) != 4:
                raise Exception("SHELL181应使用4个Node.")
            else:
                self.nlist = nodes
        elif etype == 1841 or etype == 1840:
            if len(nodes) != 2:
                raise Exception("MPC184应使用2个Node.")
            else:
                self.nlist = nodes
        else:
            raise Exception("单元类型不支持.")

    @property
    def location(self):
        x0 = np.mean([n.x for n in self.nlist])
        y0 = np.mean([n.y for n in self.nlist])
        z0 = np.mean([n.z for n in self.nlist])
        return x0, y0, z0

    @property
    def apdl_str(self):
        cmd_str = '''mat,%i
type,%i
real,%i
secn,%i
e,''' % (self.mat, self.etype, self.real, self.secn)
        for nn in self.nlist:
            cmd_str += "%i" % nn.id
            cmd_str += ','
        # cmd_str += '9999999'
        return cmd_str

    def __str__(self):
        ss = "Elem: (%i,type=%i,secn=%i,nlist=[" % (self.e, self.etype, self.secn,)
        for n in self.nlist:
            ss += "%i," % n.id
        ss += "])"
        return ss



import numpy as np
import pandas as pd
from ezdxf.math import Vec3

from .fem import FEM
from ..fem_base import Node, Element, Fix


class TrussUni(FEM):
    def __init__(self, slope, cell_l, cell_h_fuc, cell_w_fuc, pier_loc):
        super().__init__()
        self.slope = slope
        self.n0 = 0
        self.e2windy = []
        self.e2rq = []
        self.e2dead2 = []
        self.cell_l = cell_l
        self.cell_h = cell_h_fuc
        self.cell_w = cell_w_fuc
        self.pier_loc = pier_loc
        # self.make_sections()
        nodes, elements = self.make_fem()
        for i, row in nodes.iterrows():
            self.node_list[int(row['n'])] = Node(row['n'], row.x, row.y, row.z)
        for i, row in elements.iterrows():
            self.elem_list[int(row['e'])] = (
                Element(e=row.e, mat=row.mat, etype=188, real=188, secn=row.secn, nodes=
                [
                    self.node_list[row.ni],
                    self.node_list[row.nj]
                ]))
            if row.secn in [11]:
                self.top_elem_list[int(row['e'])] = (
                    Element(e=row.e, mat=row.mat, etype=188, real=188, secn=row.secn, nodes=
                    [
                        self.node_list[row.ni],
                        self.node_list[row.nj]
                    ]))
                if row['e'] // 1000 == 53:
                    self.left_elem_list[int(row['e'])] = (
                        Element(e=row.e, mat=row.mat, etype=188, real=188, secn=row.secn, nodes=
                        [
                            self.node_list[row.ni],
                            self.node_list[row.nj]
                        ]))
            elif row.secn in [12]:
                if row['e'] // 1000 == 54:
                    self.left_elem_list[int(row['e'])] = (
                        Element(e=row.e, mat=row.mat, etype=188, real=188, secn=row.secn, nodes=
                        [
                            self.node_list[row.ni],
                            self.node_list[row.nj]
                        ]))

        self.fix_list[2000] = Fix(2000, {'Uy': 0, 'Uz': 0})
        self.fix_list[4000] = Fix(4000, {'Uy': 0, 'Uz': 0})

    def make_fem(self):
        xlist = []
        zlist = []
        x0 = 0
        for i in range(len(self.cell_l)):
            xlist.append(x0)
            zlist.append(x0 * self.slope)
            x0 += self.cell_l[i]
        xlist.append(x0)
        zlist.append(x0 * self.slope)
        nodes = []
        elems = []
        nodes = pd.DataFrame(nodes, columns=['n', 'x', 'y', 'z'], dtype=np.float32)
        elems = pd.DataFrame(elems, columns=['e', 'ni', 'nj', 'secn', 'mat'], dtype=np.float32)
        for i, x in enumerate(xlist):
            for u in range(2):
                y = self.cell_w(x) * [-0.5, 0.5][u]
                for v, z in enumerate([zlist[i], zlist[i] - self.cell_h(x)]):
                    layer = int("%i%i" % (u, v), 2) + 1
                    node_id: int = self.n0 + i + layer * 1000
                    nodes.loc[len(nodes)] = [node_id, x, y, z]
            if self.inside(self.pier_loc, x - 1500, [-13600, 13600]):
                coord_secn = 11
                cell_secn = [12, 13, 14]
            else:
                coord_secn = 11
                cell_secn = [12, 13, 14]
            if self.inside(self.pier_loc, x, [-13600, 13600]):
                rect_secn = [12, 13, 14]
            else:
                rect_secn = [12, 13, 14]
            elems.loc[len(elems)] = [self.n0 + 10000 + i, self.n0 + 1000 + i, self.n0 + 2000 + i, 23, 2]
            elems.loc[len(elems)] = [self.n0 + 20000 + i, self.n0 + 2000 + i, self.n0 + 4000 + i, 22, 2]
            elems.loc[len(elems)] = [self.n0 + 30000 + i, self.n0 + 4000 + i, self.n0 + 3000 + i, 23, 2]
            elems.loc[len(elems)] = [self.n0 + 40000 + i, self.n0 + 3000 + i, self.n0 + 1000 + i, 21, 2]
            if i != 0:
                elems.loc[len(elems)] = [self.n0 + i + 51000, self.n0 + 1000 + i, self.n0 + 1000 + (i - 1), 11, 1]
                elems.loc[len(elems)] = [self.n0 + i + 52000, self.n0 + 2000 + i, self.n0 + 2000 + (i - 1), 12, 1]
                elems.loc[len(elems)] = [self.n0 + i + 53000, self.n0 + 3000 + i, self.n0 + 3000 + (i - 1), 11, 1]
                elems.loc[len(elems)] = [self.n0 + i + 54000, self.n0 + 4000 + i, self.n0 + 4000 + (i - 1), 12, 1]
                self.e2rq.append(self.get_e_loc(self.n0 + i + 53000, nodes, elems))
                self.e2dead2.append(self.get_e_loc(self.n0 + i + 51000, nodes, elems))
                self.e2dead2.append(self.get_e_loc(self.n0 + i + 53000, nodes, elems))
                self.e2windy.append(self.get_e_loc(self.n0 + i + 53000, nodes, elems))
                self.e2windy.append(self.get_e_loc(self.n0 + i + 54000, nodes, elems))
                if i % 2 == 0:
                    elems.loc[len(elems)] = [self.n0 + i + 55000, self.n0 + 1000 + i, self.n0 + 2000 + (i - 1), 23, 2]
                    elems.loc[len(elems)] = [self.n0 + i + 56000, self.n0 + 3000 + i, self.n0 + 4000 + (i - 1), 23, 2]
                    elems.loc[len(elems)] = [self.n0 + i + 57000, self.n0 + 3000 + i, self.n0 + 1000 + (i - 1), 21, 2]
                    elems.loc[len(elems)] = [self.n0 + i + 58000, self.n0 + 4000 + i, self.n0 + 2000 + (i - 1), 22, 2]
                else:
                    elems.loc[len(elems)] = [self.n0 + i + 55000, self.n0 + 2000 + i, self.n0 + 1000 + (i - 1), 23, 2]
                    elems.loc[len(elems)] = [self.n0 + i + 56000, self.n0 + 4000 + i, self.n0 + 3000 + (i - 1), 23, 2]
                    elems.loc[len(elems)] = [self.n0 + i + 57000, self.n0 + 2000 + i, self.n0 + 4000 + (i - 1), 21, 2]
                    elems.loc[len(elems)] = [self.n0 + i + 58000, self.n0 + 1000 + i, self.n0 + 3000 + (i - 1), 22, 2]
        return nodes, elems

    @staticmethod
    def inside(locs, x0, ranges):
        for lc in locs:
            if (x0 - lc >= ranges[0]) and (x0 - lc <= ranges[1]):
                return True
        return False

    @staticmethod
    def get_e_loc(eid, nodes, elems):
        n1 = elems[elems['e'] == eid]['ni'].values[0]
        n2 = elems[elems['e'] == eid]['nj'].values[0]
        n1 = nodes[nodes['n'] == n1]
        n2 = nodes[nodes['n'] == n2]
        p1 = Vec3(float(n1.x.values[0]), float(n1.y.values[0]), float(n1.z.values[0]))
        p2 = Vec3(float(n2.x.values[0]), float(n2.y.values[0]), float(n2.z.values[0]))
        return eid, 0.5 * (p1 + p2)

#     def make_sections(self):
#         self.sect_list = {
#             11: BoxSection(Id=11, Name="上弦杆", offset=(0, 0), w1=400, w2=250, t1=12, t2=12, t3=12, t4=12),
#             12: BoxSection(Id=12, Name="下弦杆", offset=(0, 0), w1=400, w2=250, t1=12, t2=12, t3=12, t4=12),
#             21: ISection(Id=21, Name="上平联", offset=(0, 200), w1=194, w2=150, w3=150, t1=9, t2=9, t3=6),
#             22: ISection(Id=22, Name="下平联", offset=(0, 200), w1=194, w2=150, w3=150, t1=9, t2=9, t3=6),
#             23: ISection(Id=23, Name="腹杆", offset=(0, 0), w1=194, w2=150, w3=150, t1=9, t2=9, t3=6),
#             31: BoxSection(Id=31, Name="墩顶弦杆", offset=(0, 0), w1=400, w2=400, t1=20, t2=20, t3=20, t4=20),
#             32: BoxSection(Id=32, Name="墩顶上平联", offset=(0, 200), w1=350, w2=250, t1=10, t2=10, t3=10, t4=10),
#             33: BoxSection(Id=33, Name="墩顶下平联", offset=(0, 200), w1=350, w2=250, t1=10, t2=10, t3=10, t4=10),
#             34: BoxSection(Id=34, Name="墩顶腹杆", offset=(0, 0), w1=350, w2=250, t1=10, t2=10, t3=10, t4=10)
#         }
#


from .base import ApdlWriteable


class Fix(ApdlWriteable):
    __slots__ = ('id', 'fix_dict')

    def __init__(self, idx, f_dict):
        self.id = idx
        self.fix_dict = f_dict

    @property
    def apdl_str(self):
        cmd_str = ""
        for key in self.fix_dict.keys():
            cmd_str += "d,%i,%s,%i\n" % (self.id, key, self.fix_dict[key])
        return cmd_str

    @property
    def mct_str(self):
        if 'all' in self.fix_dict.keys():
            cmd = f"{self.id},111111,\n"
        else:
            ux = 1 if 'Ux' in self.fix_dict.keys() else 0
            uy = 1 if 'Uy' in self.fix_dict.keys() else 0
            uz = 1 if 'Uz' in self.fix_dict.keys() else 0
            rx = 1 if 'Rx' in self.fix_dict.keys() else 0
            ry = 1 if 'Ry' in self.fix_dict.keys() else 0
            rz = 1 if 'Rz' in self.fix_dict.keys() else 0
            cmd = f"{self.id},{ux}{uy}{uz}{rx}{ry}{rz},\n"
        return cmd



import json
import os
import shutil
import subprocess

import numpy as np

from .base_api import BaseAPI
from .wind_calculation import getFg, GetUd
from .ansys_tools import GetCd, GetSmax, GetDES
from ..zoo import FEM


class AnsysAPI(BaseAPI):
    def __init__(self, project, root_path, init=True):
        # self.cwd = ""
        super().__init__(project, root_path, init)
        wdr = os.path.join(root_path, self.project_name)
        if init:
            if not os.path.exists(wdr):
                os.mkdir(wdr)
            else:
                shutil.rmtree(wdr)
                os.mkdir(wdr)

    def _batch_file(self, filestream, jobname):
        cmd = rf'''set ANS_CONSEC=YES
set ANSYS_PATH="{self.options['ANSYS_PATH']}"
%ANSYS_PATH% -b -i {jobname}.inp -o {jobname}.out'''
        with open(filestream, 'w+') as fid:
            fid.write(cmd)

    def _apdl_main(self, filestream):
        cmd = f'''!=======================================================
! Modeled by Bill with Python Automation 2024
! Units: N,mm,t,s
!=======================================================
finish
/CLEAR,NOSTART
/TITLE,{self.project_name}
/input,Etype,inp
/input,Material,inp
/input,Section,inp
/input,Node,inp
/input,Element,inp
/input,Boundary,inp
/prep7 
nsel,all
nummrg,node,1
allsel
save,fem,db
finish
'''
        if self.options['image']:
            cmd += "/input,render,inp\n"

        with open(filestream, 'w+') as fid:
            fid.write(cmd)

    @staticmethod
    def _apdl_list_out(filestream, prep_dict):
        cmd = "/prep7\n"
        for k in prep_dict.keys():
            val = prep_dict[k]
            cmd += val.apdl_str
        with open(filestream, 'w+', encoding='gbk') as fid:
            fid.write(cmd)

    @staticmethod
    def _apdl_etype(filestream):
        cmd = "/prep7"
        cmd += '''! 单元        
et,188,BEAM188
keyopt,188,1,1 
keyopt,188,3,2
et,1840,MPC184,1
et,1841,MPC184,1
'''
        with open(filestream, 'w+', encoding='utf-8') as fid:
            fid.write(cmd)

    def _apdl_node(self, filestream):
        fem = self.fem
        cmd = "/prep7\n"
        for n in fem.node_list.keys():
            cmd += fem.node_list[n].apdl_str
            cmd += "\n"
        cmd += "n,9999999,0,1e15,0\n"
        cmd += "n,8888888,0,0,1e15\n"
        with open(filestream, 'w+') as fid:
            fid.write(cmd)

    def _apdl_dead_load(self, filestream):
        cmd = '''!-------------------------------------------------------
! DC
!-------------------------------------------------------
/solu
ACEL,0,9800,0
antype,0
solve
finish
'''
        with open(filestream, 'w+') as fid:
            fid.write(cmd)
            fid.write(self.write_etab(1, 'DC_SE1'))
            fid.write(self.write_etab(11, 'DC_SE11'))

    def _apdl_elem(self, filestream):
        fem = self.fem
        cmd = "/prep7\n"
        for e in fem.elem_list.keys():
            cmd += fem.elem_list[e].apdl_str
            if fem.elem_list[e].secn in [11, 12, 21, 22]:
                cmd += '9999999'
            cmd += "\n"
        # cmd += "esel,s,secn,,2\n"
        # cmd += "cm,girder,elem\n"
        # cmd += "nsle,s,1\n"
        # cmd += "nsel,u,node,,999999\n"
        # cmd += "cm,gnode,node\n"
        # cmd += "allsel\n"
        with open(filestream, 'w+') as fid:
            fid.write(cmd)

    @staticmethod
    def _apdl_render(filestream):
        cmd = f'''/SHOW,png,,
/DEVICE,PNG
/VIEW,1,0,0,1       
/ANG,1,X,90         
/ERASE              
/PREP7
/ESHAPE,1    
/PBC,ALL, ,1       
EPLOT               
/SHOW,CLOSE         
/DEVICE,CLOSE'''
        with open(filestream, 'w+', encoding='utf-8') as fid:
            fid.write(cmd)

    def run_fem(self, fem_model: FEM):
        self.fem = fem_model
        self._apdl_etype(os.path.join(self.cwd, 'etype.inp'))
        self._apdl_list_out(os.path.join(self.cwd, 'Material.inp'), fem_model.mat_list)
        self._apdl_list_out(os.path.join(self.cwd, 'Section.inp'), fem_model.sect_list)
        self._apdl_node(os.path.join(self.cwd, 'Node.inp'))
        self._apdl_elem(os.path.join(self.cwd, 'Element.inp'), )
        self._apdl_list_out(os.path.join(self.cwd, 'Boundary.inp'), fem_model.fix_list)
        self._apdl_render(os.path.join(str(self.cwd), 'render.inp'))
        self._apdl_main(os.path.join(str(self.cwd), 'main.inp'))
        self._batch_file(os.path.join(self.cwd, 'run.bat'), 'main')
        if not self.options['only_cmd']:
            subprocess.call(os.path.join(self.cwd, 'run.bat'), shell=True, cwd=self.cwd,
                            stdout=subprocess.DEVNULL,  # 抑制标准输出
                            stderr=subprocess.DEVNULL  # 抑制标准错误输出
                            )
            # print(f'{self.project_name}')

    def _run_ansys(self, bat_file):
        subprocess.call(os.path.join(self.cwd, bat_file), shell=True, cwd=self.cwd,
                        stdout=subprocess.DEVNULL,  # 抑制标准输出
                        stderr=subprocess.DEVNULL  # 抑制标准错误输出
                        )
        # print(bat_file)

    @staticmethod
    def write_etab(secn, tab_name, set_id='Last'):
        cmd = f'''/post1
    set,{set_id}
    esel,s,secn,,{secn}
    ETABLE, {tab_name}, SMISC, 31
    PRETAB,{tab_name}   
    finish
    '''
        return cmd

    def run_rs(self, case_name, lcn, direction, Ci, Cs, A, Tg, dr=0.03):
        """
        写入反应谱计算结果
        """
        Cd0 = GetCd(0.03)
        Smax_x = GetSmax(Ci, Cs, Cd0, A)
        T, DES_x = GetDES(Smax_x, Tg, 0.1, 10, 0.1)
        Freq = 1 / np.array(T)
        Acc = np.array(DES_x) * 9806
        Freq = np.flip(Freq)
        Acc = np.flip(Acc)
        sfreq = ["%.5f" % a for a in Freq]
        sacel = ["%.5f" % a for a in Acc]
        if direction == "X":
            sdir = '1,0,0'
        elif direction == "Y":
            sdir = '0,1,0'
        else:
            sdir = '0,0,1'
        cmd = f'''/clear,nostart 
!/FILNAME,mode,0
/solution
resume,mode,db    
antype,8               
SPOPT,SPRS,,1,1
dmprat,0.03                
grp,0.001                   
svtyp,2   
sed,{sdir}
'''
        for line in str2lines(sfreq, 9):
            cmd += "FREQ," + line + '\n'
        for line in str2lines(sacel, 9):
            cmd += "SV,%.2f," % dr + line + '\n'
        cmd += f'''
solve 
finish 
/post1
/input,file,mcom
set,last
lcdef,1
lcwrite,{lcn}
'''
        with open(os.path.join(self.cwd, f'rs_{case_name}.inp'), 'w+', encoding='utf-8') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'run_{case_name}.bat'), f'rs_{case_name}')
            self._run_ansys(f'run_{case_name}.bat')

    def run_mode(self, nmode: int):
        cmd = f'''/clear,nostart 
!/filename,model,0
/solution
resume,fem,db
ANTYPE,2
MODOPT,LANB,{nmode}  
EQSLV,SPAR  
MXPAND,{nmode}  , , ,1 
LUMPM,0 
PSTRES,0
MODOPT,LANB,{nmode}  ,0,0, ,OFF
SOLVE   
save,mode,db    
FINISH  '''
        with open(os.path.join(self.cwd, 'model.inp'), 'w+', encoding='utf-8') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, 'run_model.bat'), 'model')
            self._run_ansys('run_model.bat')
            with open(os.path.join(self.cwd, 'model.out'), 'r') as fid:
                lines = fid.readlines()
            idx = []
            for jj, l in enumerate(lines):
                if l.__contains__("   TOTAL MASS"):
                    with open(self.res_dir, 'a+') as fid:
                        fid.write("\n" + str(float(l.split()[-1])))
                if l.startswith("   sum                        "):
                    idx.append(jj)
                if l.startswith("  MODE    FREQUENCY (HERTZ)  "):
                    idx.append(jj)
                    idx.append(jj + 3)
            with open(self.res_dir, 'a+') as fid:
                fid.write("\n" + lines[idx[0]])
                fid.write(lines[idx[1]])
                fid.write(lines[idx[-1]])
            return

    def run_deadload(self, lcn=1):
        cmd = f'''/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,9800,0
allsel
solve
save,dc,db   
finish
/post1
set,last
lcdef,1
lcwrite,{lcn}
        '''
        with open(os.path.join(self.cwd, 'dc.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, 'run_dc.bat'), 'dc')
            self._run_ansys('run_dc.bat')

    def run_dw(self, f2=1.6, lcn=2):
        cmd = f'''/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,0,0
esel,s,secn,,11
sfbeam,all,2,pres,-{f2 * 0.5}
allsel
solve
finish
/post1
set,last
lcdef,1
lcwrite,{lcn}
            '''
        with open(os.path.join(self.cwd, 'dw.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, 'run_dw.bat'), 'dw')
            self._run_ansys('run_dw.bat')

    def run_rq(self, rq, lcn):
        cmd = f'''/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,0,0
esel,s,secn,,11
sfbeam,all,2,pres,-{rq * 0.5}
allsel
solve
finish
/post1
set,last
lcdef,1
lcwrite,{lcn}
                '''
        with open(os.path.join(self.cwd, 'rq.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, 'run_rq.bat'), 'rq')
            self._run_ansys('run_rq.bat')

    def run_wind(self, case_name, wind_speed, tower_locs, tower_z0, truss_z0, truss_height, lcn):
        cmd = '''/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,0,0
'''
        Ud = GetUd(wind_speed, truss_z0, 0.785, 1, 0.22).m
        Fg = getFg(1.2, Ud, 0.55, truss_height).m * 1e-3  # N/mm
        cmd += f'''esel,s,secn,,11
esel,r,cent,z,-10000,0
sfbeam,all,2,pres,-{Fg}
'''
        sps = np.mean([tower_locs[i + 1] - tower_locs[i] for i in range(len(tower_locs) - 1)])
        for tx, tz in zip(tower_locs, tower_z0):
            Ud = GetUd(wind_speed, tz, 0.785, 1, 0.22).m
            Fg = getFg(1.2, Ud, 0.55, truss_height).m * 1e-3  # N/mm
            cmd += f"esel,s,secn,,1\n"
            cmd += f"esel,r,cent,z,-10000,0\n"
            cmd += f"esel,r,cent,x,{tx - 0.5 * sps},{tx + 0.5 * sps}\n"
            cmd += f"sfbeam,all,1,pres,-{Fg}\n"
        cmd += f'''allsel
solve
save,wind_{case_name},db   
finish
/post1
set,last
lcdef,2
lcwrite,{lcn}
'''
        with open(os.path.join(self.cwd, f'wind_{case_name}.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'run_wind_{case_name}.bat'), f'wind_{case_name}')
            self._run_ansys(f'run_wind_{case_name}.bat')

    def get_node_deformation(self, config: dict):
        cmd = f'''/clear,nostart 
/post1
resume,fem,db
'''
        for key in config.keys():
            lc = config[key][0]
            direction = config[key][1]
            var_name = config[key][2]
            cmd += f'''
lcfile,{lc}
lcoper,zero
FORCE,STATIC 
lcase,{lc}
NSORT, U, {direction}, 0, 1,3
*GET, {var_name}, SORT, 0, Max
'''

        with open(os.path.join(self.cwd, 'nodel_def.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, 'nodel_def.bat'), 'nodel_def')
            self._run_ansys('nodel_def.bat')
            with open(os.path.join(self.cwd, 'nodel_def.out'), 'r') as fid:
                lines = fid.readlines()
            res = {}
            for key in config.keys():
                var_name = config[key][2]
                for line in lines:
                    if line.__contains__(f"{var_name.upper()}") and line.__contains__(" *GET"):
                        vals = line.split()
                        res[key] = float(vals[-1])
            with open(self.res_dir, 'a+') as fid:
                fid.write(json.dumps(res, ensure_ascii=False, indent=4))
            return

    def get_elem_forces(self, secn, lcn):
        cmd = f'''
/clear,nostart 
/post1
resume,fem,db   
lcfile,{lcn}
esel,s,secn,,{secn}
lcoper,zero
FORCE,STATIC 
lcase,{lcn}
ETABLE, sxx, SMISC, 31
ETABLE, syT, SMISC, 32
ETABLE, syB, SMISC, 33
ETABLE, szT, SMISC, 34
ETABLE, szB, SMISC, 35
sadd,sTT,sxx,syT
sadd,sTT,sTT,szT
sadd,sTB,sxx,syT
sadd,sTB,sTB,szB
sadd,sBB,sxx,syB
sadd,sBB,sBB,szB
sadd,sBT,sxx,syB
sadd,sBT,sBT,szT
SMAX,smaxT,sTT,sTB
SMAX,smaxB,sBT,sBB
SMAX,smax,smaxT,smaxB
SMIN,sminT,sTT,sTB
SMIN,sminB,sBT,sBB
SMIN,smin,sminT,sminB
ESORT, ETAB, smin, 1,0,1
ESORT, ETAB, smax, 0,0,1
PRETAB, smin, smax
'''
        with open(os.path.join(self.cwd, f'ET_{secn}_{lcn}.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'et_{secn}_{lcn}.bat'), f'et_{secn}_{lcn}')
            self._run_ansys(f'et_{secn}_{lcn}.bat')
            with open(os.path.join(self.cwd, f'et_{secn}_{lcn}.out'), 'r') as fid:
                lines = fid.readlines()
            res = {}
            for ii, line in enumerate(lines):
                if line.__contains__(" MINIMUM VALUES"):
                    vals = [float(a) for a in lines[ii + 2].split()[-2:]]
                    res[f'{lcn}_{secn}_min'] = vals[0]
                    res[f'{lcn}_{secn}_max'] = vals[1]
            with open(self.res_dir, 'a+') as fid:
                fid.write('\n')
                fid.write(json.dumps(res, ensure_ascii=False, indent=4))
            return

    def srss_rs(self, case_name, lcn, lc_x, lc_y, lc_z):
        cmd = f'''/clear,nostart 
/post1
resume,fem,db    
lcfile,{lc_x}
lcfile,{lc_y}
lcfile,{lc_z}
lcoper,zero
FORCE,STATIC 
lcase,{lc_x}
lcoper,SQUA
LCOPER,ADD,{lc_y},MULT,{lc_y}
LCOPER,ADD,{lc_z},MULT,{lc_z}
LCOPER,SQRT
lcwrite,{lcn}
'''
        with open(os.path.join(self.cwd, f'{case_name}.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'run_{case_name}.bat'), case_name)
            self._run_ansys(f'run_{case_name}.bat')

    def run_lcomb(self, case_name, conf):
        cmd = ""
        for key in conf.keys():
            cmd += f'''/clear,nostart 
/post1
resume,fem,db    
'''
            lc = conf[key]
            for sub_lc, fct in lc:
                cmd += f"lcfile,{sub_lc}\n"
                cmd += f"lcfact,{sub_lc},{fct}\n"
            cmd += f''' 
lcoper,zero
FORCE,STATIC 
lcase,{lc[0][0]}
'''
            for sub_lc, fct in lc[1:]:
                cmd += f"lcoper,add,{sub_lc}\n"
            cmd += f"lcwrite,{key}\n"
        with open(os.path.join(self.cwd, f'{case_name}.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'run_{case_name}.bat'), case_name)
            self._run_ansys(f'run_{case_name}.bat')

    def run_tu(self, case_name, temp_up, temp_down, lcn):
        cmd = f'''/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,0,0
TUNIF, {temp_up}
allsel
solve
finish
/post1
set,last
lcdef,1
lcwrite,{lcn + 1}
/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,0,0
TUNIF, {temp_down}
allsel
solve
finish
/post1
set,last
lcdef,1
lcwrite,{lcn + 2}
'''
        with open(os.path.join(self.cwd, f'{case_name}.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'run_{case_name}.bat'), case_name)
            self._run_ansys(f'run_{case_name}.bat')

    def run_se(self, case_name, num_tower, val, lcn):
        fd = np.array(list(self.fem.fix_list.keys()))
        cmd = ""
        for i in range(num_tower):
            cmd += f'''/clear,nostart 
/solution
resume,fem,db
antype,0
ACEL,0,0,0
'''
            cur_n = fd[np.where(fd // 1e6 == (i + 1))]
            for n in cur_n:
                cmd += f"d,{n},uy,-{val}\n"
            cmd += f'''allsel
solve
finish
/post1
set,last
lcdef,1
lcwrite,{51 + i}
'''
        cmd += f'''/clear,nostart 
/post1
resume,fem,db   
'''
        for i in range(num_tower):
            cmd += f"lcfile,{51 + i}\n"
        cmd += '''lcoper,zero
FORCE,STATIC 
lcase,51
        '''
        for i in range(num_tower - 1):
            cmd += f"lcoper,ABM X,{52 + i}\n"
        cmd += f"lcwrite,{lcn}\n"
        with open(os.path.join(self.cwd, f'{case_name}.inp'), 'w+') as fid:
            fid.write(cmd)
        if not self.options['only_cmd']:
            self._batch_file(os.path.join(self.cwd, f'run_{case_name}.bat'), case_name)
            self._run_ansys(f'run_{case_name}.bat')

    def clear(self):
        if self.init:
            shutil.rmtree(self.cwd)
        pass

    def init_records(self, dimensions, loads):
        with open(self.res_dir, 'w+') as fid:
            json_string = json.dumps(dimensions, ensure_ascii=False, indent=4)
            fid.write(json_string)
            fid.write("\n")
            json_string = json.dumps(loads, ensure_ascii=False, indent=4)
            fid.write(json_string)


def str2lines(str_list, n):
    lines = [",".join(str_list[i:i + n]) for i in range(0, len(str_list), n)]
    return lines



from typing import Union

from .fem import FEM
from ..fem_base import Material
from ..fem_base.section import Section


class SectionGroup(FEM):
    def __init__(self):
        super().__init__()

    def add(self, obj: Union[Section, Material]):
        if isinstance(obj, Section):
            self.sect_list[obj.id] = obj
        elif isinstance(obj, Material):
            self.mat_list[obj.id] = obj
        else:
            raise TypeError



from ezdxf.math import Vec3

from .base import ApdlWriteable


class Node(ApdlWriteable):
    __slots__ = ('id', 'x', 'y', 'z', 'vec')

    def __init__(self, n: int, *args):
        self.id = n
        args_list = list(args)
        if len(args_list) == 3:
            self.vec: Vec3 = Vec3(args_list)
            self.x = float(args_list[0])
            self.y = float(args_list[1])
            self.z = float(args_list[2])
        elif len(args_list) == 1:
            self.vec: Vec3 = args_list[0]
            self.x = self.vec.x
            self.y = self.vec.y
            self.z = self.vec.z
        else:
            raise ValueError

    @property
    def apdl_str(self):
        cmd_str = "n,%i,%.3f,%.3f,%.3f" % (self.id, self.x, self.z, self.y)
        return cmd_str

    def __str__(self):
        return "Node: (%i,%.3f,%.3f,%.3f)" % (self.id, self.x, self.y, self.z)

    def distance(self, other: 'Node'):
        return self.vec.distance(other.vec)

    def copy(self, dn: int = 0, x0: float = 0, y0: float = 0, z0: float = 0):
        return Node(self.n + dn, self.x + x0, self.y + y0, self.z + z0)

    def __eq__(self, other: 'Node'):
        return self.n == other.n

    def __ne__(self, other: 'Node'):
        return self.n != other.n

    def __gt__(self, other: 'Node'):
        return self.n > other.n

    def __lt__(self, other: 'Node'):
        return self.n < other.n

    def __ge__(self, other: 'Node'):
        return self.n >= other.n

    def __le__(self, other: 'Node'):
        return self.n <= other.n


if __name__ == '__main__':
    n1 = Node(1, 1, 1, 1)



from .base import ApdlWriteable


class Material(ApdlWriteable):
    __slots__ = ('id', 'ex', 'dens', 'alpx', 'nuxy', 'description')

    def __init__(self, id, ex, dens, alpx, nuxy, description: str = ''):
        self.id = id
        self.ex = ex
        self.dens = dens
        self.alpx = alpx
        self.nuxy = nuxy
        self.description = description

    @property
    def mct_str(self):
        cmd = f" {self.id}, USER , {self.description} , 0.06, 0, , C, NO, 0, 2, 2.0600e+005,"
        cmd += f"{self.nuxy}, {self.alpx}, {self.dens * 9806},     0"
        return cmd

    @property
    def apdl_str(self):
        cmd_str = '''
!%s
MP,EX,  %i,%.4e
MP,DENS,%i,%.4e
MP,ALPX,%i,%.4e
MP,NUXY,%i,%.1f''' % (self.description,
                      self.id, self.ex,
                      self.id, self.dens,
                      self.id, self.alpx,
                      self.id, self.nuxy
                      )
        return cmd_str



import os
import shutil

import numpy as np

from .base_api import BaseAPI
from ..zoo import FEM


class MidasDRS:
    def __init__(self, bridge_type, site_type, intensity, Tg, kesi, seismic_type, is_vertical, Tmax, delt_t):
        """

        :param bridge_type: 桥梁类别
        :param site_type: 场地类别（"I0", "I1", "II", "III", "IV"）
        :param intensity:  设防烈度（"VI", "VII", "VII+", "VIII", "VIII+", "IX", ）
        :param Tg:  分区特征周期 （0.35, 0.4, 0.45）
        :param kesi: 阻尼比
        :param seismic_type: 反应谱类别（"E1","E2"）
        :param is_vertical: 是否为竖向地震
        :param Tmax: 生成长度
        """
        self.specification = "JTG/T 2231-01-2020"

        assert Tg in [0.35, 0.4, 0.45]
        assert site_type in ["I0", "I1", "II", "III", "IV"]
        assert intensity in ["VI", "VII", "VII+", "VIII", "VIII+", "IX", ]
        self.Tg = Tg
        self.bridge_type = bridge_type
        self.seismic_type = seismic_type
        self.intensity = intensity
        self.damper_ratio = kesi
        self.site_type = site_type
        self.isVertical = is_vertical
        self.A = self.get_A(intensity)
        self.Ci = self.get_Ci(seismic_type, bridge_type)
        self.Cd = self.get_Cd(kesi)
        self.Cs = self.get_Cs(site_type, intensity, is_vertical)
        self.Tg_real = self.get_Tgft(site_type, Tg, is_vertical)
        self.Smax = self.get_Smax(self.Ci, self.Cs, self.Cd, self.A)
        self.Tmax = Tmax
        self.ts, self.ss = self.get_DES(delt_t)
        self.parameters = self.get_para_mct()

    def get_para_mct(self):

        if self.site_type == "I0":
            st = 1
        elif self.site_type == "I1":
            st = 2
        elif self.site_type == "II":
            st = 3
        elif self.site_type == "III":
            st = 4
        elif self.site_type == "IV":
            st = 5
        else:
            raise KeyError

        if self.intensity == "VI":
            it = "6(0.05g)"
        elif self.intensity == "VII":
            it = "7(0.1g)"
        elif self.intensity == "VII+":
            it = "7(0.15g)"
        elif self.intensity == "VIII":
            it = "8(0.2g)"
        elif self.intensity == "VIII+":
            it = "8(0.3g)"
        elif self.intensity == "IX":
            it = "9(0.4g)"
        else:
            raise KeyError
        isv = 1 if self.isVertical else 0
        return ("   %s, %s, %.2f, %i, %s, %s,%.2f, %.2f, %.2f, %.2f, %.2f, %.8f, %.2f,%.2f, 0, %i\n"
                % (
                    self.specification,
                    self.bridge_type,
                    self.Tg,
                    st, it, self.seismic_type, self.Tg_real, self.Ci, self.Cs, self.Cd, self.A, self.Smax,
                    self.Tmax, self.damper_ratio, isv
                )
                )

    def get_DES(self, step):
        Tlist = np.linspace(0, self.Tmax, int(np.round((self.Tmax - 0) / step + 1))).tolist()
        Tlist.append(0.1)
        Tlist.append(self.Tg_real)
        Tlist = list(set(Tlist))
        Tlist.sort()
        Slist = [self.get_S(self.Smax, self.Tg_real, ti) for ti in Tlist]
        return Tlist, Slist

    @staticmethod
    def get_S(Smax, Tg, T):
        T0 = 0.1
        if T < T0:
            return Smax * (0.6 * T / T0 + 0.4)
        elif T < Tg:
            return Smax
        else:
            return Smax * (Tg / T)

    @staticmethod
    def get_Smax(Ci, Cs, Cd, A):
        return 2.5 * Ci * Cs * Cd * A

    @staticmethod
    def get_Cs(site_type, intensity, isV):
        res = [
            [0.72, 0.74, 0.75, 0.76, 0.85, 0.9],
            [0.80, 0.82, 0.83, 0.85, 0.95, 1.0],
            [1.00, 1.00, 1.00, 1.00, 1.00, 1.0],
            [1.30, 1.25, 1.15, 1.00, 1.00, 1.0],
            [1.25, 1.20, 1.10, 1.00, 0.95, 0.9],
        ]
        resV = [
            [0.6, 0.6, 0.6, 0.6, 0.6, 0.6],
            [0.6, 0.6, 0.6, 0.6, 0.7, 0.7],
            [0.6, 0.6, 0.6, 0.6, 0.7, 0.8],
            [0.7, 0.7, 0.7, 0.8, 0.8, 0.8],
            [0.8, 0.8, 0.8, 0.9, 0.9, 0.8],
        ]
        ridx = ["I0", "I1", "II", "III", "IV"].index(site_type)
        cidx = ["VI", "VII", "VII+", "VIII", "VIII+", "IX", ].index(intensity)
        if isV:
            return resV[ridx][cidx]
        else:
            return res[ridx][cidx]

    @staticmethod
    def get_Tgft(site_type, Tg, isV):
        if not isV:
            if Tg == 0.35:
                if site_type == "I0":
                    return 0.20
                elif site_type == "I1":
                    return 0.25
                elif site_type == "II":
                    return 0.35
                elif site_type == "III":
                    return 0.45
                elif site_type == "IV":
                    return 0.65
                else:
                    raise KeyError
            elif Tg == 0.40:
                if site_type == "I0":
                    return 0.25
                elif site_type == "I1":
                    return 0.30
                elif site_type == "II":
                    return 0.40
                elif site_type == "III":
                    return 0.55
                elif site_type == "IV":
                    return 0.75
                else:
                    raise KeyError
            elif Tg == 0.45:
                if site_type == "I0":
                    return 0.30
                elif site_type == "I1":
                    return 0.35
                elif site_type == "II":
                    return 0.45
                elif site_type == "III":
                    return 0.65
                elif site_type == "IV":
                    return 0.90
                else:
                    raise KeyError(site_type)
        else:
            if Tg == 0.35:
                if site_type == "I0":
                    return 0.15
                elif site_type == "I1":
                    return 0.20
                elif site_type == "II":
                    return 0.25
                elif site_type == "III":
                    return 0.30
                elif site_type == "IV":
                    return 0.55
                else:
                    raise KeyError
            elif Tg == 0.40:
                if site_type == "I0":
                    return 0.20
                elif site_type == "I1":
                    return 0.25
                elif site_type == "II":
                    return 0.30
                elif site_type == "III":
                    return 0.35
                elif site_type == "IV":
                    return 0.60
                else:
                    raise KeyError
            elif Tg == 0.45:
                if site_type == "I0":
                    return 0.25
                elif site_type == "I1":
                    return 0.30
                elif site_type == "II":
                    return 0.40
                elif site_type == "III":
                    return 0.50
                elif site_type == "IV":
                    return 0.75
                else:
                    raise KeyError

    @staticmethod
    def get_Cd(xi):
        return max(1 + (0.05 - xi) / (0.08 + 1.6 * xi), 0.55)

    @staticmethod
    def get_Ci(seismic_type, bridge_type):
        if seismic_type == "E1":
            if bridge_type == "A":
                return 1.0
            elif bridge_type == "B":
                return 0.43
            elif bridge_type == "C":
                return 0.34
            elif bridge_type == "D":
                return 0.23
            else:
                raise KeyError
        elif seismic_type == "E2":
            if bridge_type == "A":
                return 1.7
            elif bridge_type == "B":
                return 1.3
            elif bridge_type == "C":
                return 1.0
            elif bridge_type == "D":
                raise KeyError
            else:
                raise KeyError
        else:
            raise KeyError
        pass

    @staticmethod
    def get_A(intensity):
        if intensity == "VI":
            return 0.05
        elif intensity == "VII":
            return 0.10
        elif intensity == "VII+":
            return 0.15
        elif intensity == "VIII":
            return 0.20
        elif intensity == "VIII+":
            return 0.30
        elif intensity == "IX":
            return 0.40
        else:
            raise KeyError


class MidasAPI(BaseAPI):

    def __init__(self, project, root_path, init=True):
        super().__init__(project, root_path, init)
        self.options = {}
        self.fem: 'FEM' = None
        self.project_name = project
        wdr = os.path.join(root_path, self.project_name)
        self.cwd = os.path.abspath(wdr)
        self.res_dir = os.path.join(os.path.abspath(root_path), f'{project}_result.txt')
        self.init = init
        if init:
            if not os.path.exists(wdr):
                os.mkdir(wdr)
            else:
                shutil.rmtree(wdr)
                os.mkdir(wdr)

    def run_fem(self, fem_model: FEM):
        self.fem = fem_model
        file_out = open(os.path.join(self.cwd, self.project_name + '.mct'), 'w+', encoding='GBK')
        self._begin(file_out)
        self._temp(file_out)
        self._sections(file_out)
        self._nodes(file_out)
        self._elems(file_out)
        self._constraint(file_out)
        self.to_mct_rq(file_out, self.loads['rq'])
        self.to_mct_dead2(file_out, self.loads['dw'], self.loads['snow'])
        self.to_mct_wind(file_out, 20, 40, 0.785, 0.22, 1.26, 0.87, 40, truss_height_m=3.0)
        if self.loads.keys().__contains__("e1"):
            self.to_mct_seismic(file_out, self.loads['e1'], self.loads['e2'])
        file_out.close()

    def _begin(self, fid):
        fid.write("*UNIT\n")
        fid.write("N, mm, KJ, C\n")
        fid.write("*STRUCTYPE\n")
        fid.write("0,1,1,NO,YES,9806,0,NO,NO,NO\n")
        if len(self.fem.mat_list) != 0:
            fid.write("*MATERIAL \n")
            for ky in self.fem.mat_list.keys():
                m = self.fem.mat_list[ky]
                fid.write(m.mct_str)
                fid.write('\n')
        fid.write("*STLDCASE\n")
        fid.write("; LCNAME, LCTYPE, DESC\n")
        fid.write("   自重 , USER, \n")
        fid.write("   二期 , USER, \n")
        fid.write("   W1主梁横风, USER, \n")
        fid.write("   W1立柱横风, USER, \n")
        fid.write("   W1立柱纵风, USER, \n")
        fid.write("   W2主梁横风, USER, \n")
        fid.write("   W2立柱横风, USER, \n")
        fid.write("   W2立柱纵风, USER, \n")
        fid.write("   整体升温, USER, \n")
        fid.write("   整体降温, USER, \n")
        fid.write("   雪荷载, USER, \n")
        fid.write("*USE-STLD, 自重\n")
        fid.write("*SELFWEIGHT\n")
        fid.write("0, 0, -1,\n")

    def _temp(self, fid):
        fid.write("*USE-STLD, 整体升温\n")
        fid.write("*SYSTEMPER\n")
        fid.write(f"{self.loads['temp_up']}, \n")
        fid.write("*USE-STLD, 整体降温\n")
        fid.write("*SYSTEMPER\n")
        fid.write(f"{self.loads['temp_down']}, \n")

    def _sections(self, fid):
        fid.write("*SECTION\n")
        for ky in self.fem.sect_list.keys():
            s = self.fem.sect_list[ky]
            if hasattr(s, 'mct_str'):
                fid.write(s.mct_str)

    def _nodes(self, fid):
        fid.write("*NODE\n")
        for n in self.fem.node_list.keys():
            nd = self.fem.node_list[n]
            fid.write("%i,%.6f,%.6f,%.6f\n" % (nd.id, nd.x, nd.y, nd.z))

    def _elems(self, fid):
        fid.write("*ELEMENT\n")
        for ky in self.fem.elem_list.keys():
            e = self.fem.elem_list[ky]
            if len(e.nlist) == 4:
                #   5060, PLATE ,    1,     1,  7001,  7002,  2002,  2001,     1,     0
                n1 = e.nlist[0].id
                n2 = e.nlist[1].id
                n3 = e.nlist[2].id
                n4 = e.nlist[3].id
                iMat = e.mat
                iSecn = e.secn
                fid.write(" %i,PLATE,%i,%i,%i,%i,%i,%i,1,0\n" % (e.id, iMat, iSecn, n1, n2, n3, n4))
            else:
                n1 = e.nlist[0].id
                n2 = e.nlist[1].id
                iMat = e.mat
                iSecn = e.secn
                beta = 90 if (iSecn == 15 or iSecn == 17) else 0
                fid.write(" %i,BEAM,%i,%i,%i,%i,%i,0\n" % (e.id, iMat, iSecn, n1, n2, beta))

    def _constraint(self, fid):
        fid.write("*CONSTRAINT    ; Supports\n")
        for ky in self.fem.fix_list.keys():
            f = self.fem.fix_list[ky]
            if hasattr(f, 'mct_str'):
                fid.write(f.mct_str)

    def to_mct_dead2(self, fid, dw2, snow):
        e2dead2 = [e for e in self.fem.top_elem_list.keys() if e // 1000 == 51 or e // 1000 == 53]
        e2dead2.sort()
        if len(e2dead2) != 0:
            fid.write("*USE-STLD,二期\n")
            fid.write("*BEAMLOAD \n")
        for ee in e2dead2:
            fid.write(
                " %i, BEAM   , UNILOAD, GZ, NO , NO, aDir[1], , , , 0, %.3f, 1, %.3f, 0, 0, 0, 0, , NO, 0, 0, NO, \n" % (
                    ee, -dw2 * 0.5, -dw2 * 0.5))
        fid.write("*USE-STLD,雪荷载\n")
        fid.write("*BEAMLOAD \n")
        for ee in e2dead2:
            fid.write(
                " %i, BEAM   , UNILOAD, GZ, NO , NO, aDir[1], , , , 0, %.3f, 1, %.3f, 0, 0, 0, 0, , NO, 0, 0, NO, \n" % (
                    ee, -snow * 0.5, -snow * 0.5))

    def to_mct_rq(self, fid, rq):
        fid.write("*MVLDCODE\n")
        fid.write("   CODE=CHINA\n")
        fid.write("*LINELANE(CH) \n")
        fid.write("   NAME=L, LANE, , 0, 0, BOTH, 500, 3000, NO, 3000\n")
        e2rq = [e for e in self.fem.top_elem_list.keys() if e // 1000 == 53]
        e2rq.sort()
        for ii, e in enumerate(e2rq):
            if ii == 0:
                st = "YES"
                ed = ", "
            elif ii != len(e2rq) - 1:
                st = "NO"
                if ii % 2 != 0:
                    ed = '\n'
                else:
                    ed = ", "
            else:
                st = "NO"
                ed = '\n'
            fid.write("     %i, %f, 120000, %s ,1 %s" % (e, 2250, st, ed))
        fid.write("*VEHICLE\n")
        fid.write(f"   NAME=专用人行荷载, 2, CROWD, 1, {rq}\n")
        fid.write("*MVLDCASE(CH)   \n")
        fid.write("   NAME=RQ, , NO, 2, 1, 0\n")
        fid.write("        1, 1, 0.8, 0.67, 0.6, 0.55, 0.55, 0.55\n")
        fid.write("        1, 1, 0.78, 0.67, 0.6, 0.55, 0.52, 0.5\n")
        fid.write("        1.2, 1, 0.78, 0.67, 0.6, 0.55, 0.52, 0.5\n")
        fid.write("        VL, 专用人行荷载, 1, 0, 1, L\n")
        fid.write("*MOVE-CTRL(CH) \n")
        fid.write("   INF, 0, 3, NODAL, NO, AXIAL, YES,   YES, NO, ,   YES, NO, ,   YES, NO, ,   YES, NO,   , NO, 0, 0, YES, 0\n")
        fid.write("   0\n")

    # def to_mct_wind(self, fid, U10W1, U10W2, kc, alpha0, gv, CY, func_D):
    def to_mct_wind(self, fid, U10W1, U10W2, kc, alpha0, gv, CY, truss_z0_m, truss_height_m):
        zz0 = truss_z0_m
        UdW1 = self.GetUd(U10W1, zz0, kc, 1, alpha0).m
        FgW1 = self.getFg(gv, UdW1, CY, truss_height_m).m * 1e-3  # N/mm
        UdW2 = self.GetUd(U10W2, zz0, kc, 1, alpha0).m
        FgW2 = self.getFg(gv, UdW2, CY, truss_height_m).m * 1e-3  # N/mm

        e2windy_beam = [e for e in self.fem.left_elem_list.keys() if e < 1e5]  # or e // 1000 == 54]
        fid.write("*USE-STLD,W1主梁横风\n")
        fid.write("*BEAMLOAD \n")
        for ee in e2windy_beam:
            fid.write(
                " %i, BEAM , UNILOAD, GY, NO , NO, aDir[1], , , , 0, %.3f, 1, %.3f, 0, 0, 0, 0, , NO, 0, 0, NO, \n" % (
                    ee, -FgW1 * 0.5, -FgW1 * 0.5))
        fid.write("*USE-STLD,W2主梁横风\n")
        fid.write("*BEAMLOAD \n")
        for ee in e2windy_beam:
            fid.write(
                " %i, BEAM , UNILOAD, GY, NO , NO, aDir[1], , , , 0, %.3f, 1, %.3f, 0, 0, 0, 0, , NO, 0, 0, NO, \n" % (
                    ee, -FgW2 * 0.5, -FgW2 * 0.5))
        e2windx_tower = [e for e in self.fem.left_elem_list.keys() if e > 1e5]
        fid.write("*USE-STLD,W1立柱横风\n")
        fid.write("*BEAMLOAD \n")
        for ee in e2windx_tower:
            ele = self.fem.elem_list[ee]
            z0 = ele.location[0] * 1e-3 * self.options['slope'] - self.options['truss_h'] * 1e-3
            dist = z0 - ele.location[2] * 1e-3
            zi = zz0 - dist
            dx = self.options['k_l'] * dist
            D = self.options['truss_l'] * 1e-3 + 2 * dx
            zi = max(zi, 0.1)
            UdW1 = self.GetUd(U10W1, zi, kc, 1, alpha0).m
            FgW1 = self.getFg(gv, UdW1, CY, D).m * 1e-3  # N/mm
            fid.write(
                " %i, BEAM , UNILOAD, GY, NO , NO, aDir[1], , , , 0, %.3f, 1, %.3f, 0, 0, 0, 0, , NO, 0, 0, NO, \n" % (
                    ee, -FgW1 * 0.5, -FgW1 * 0.5))
        fid.write("*USE-STLD,W2立柱横风\n")
        fid.write("*BEAMLOAD \n")
        for ee in e2windx_tower:
            ele = self.fem.elem_list[ee]
            z0 = ele.location[0] * 1e-3 * self.options['slope'] - self.options['truss_h'] * 1e-3
            dist = z0 - ele.location[2] * 1e-3
            zi = zz0 - dist
            dx = self.options['k_l'] * dist
            D = self.options['truss_l'] * 1e-3 + 2 * dx
            zi = max(zi, 0.1)
            UdW2 = self.GetUd(U10W2, zi, kc, 1, alpha0).m
            FgW2 = self.getFg(gv, UdW2, CY, D).m * 1e-3  # N/mm
            fid.write(
                " %i, BEAM , UNILOAD, GY, NO , NO, aDir[1], , , , 0, %.3f, 1, %.3f, 0, 0, 0, 0, , NO, 0, 0, NO, \n" % (
                    ee, -FgW2 * 0.5, -FgW2 * 0.5))
        return

    def to_mct_seismic(self, fid, spa_e1: MidasDRS, spa_e2: MidasDRS):
        fid.write("*UNIT\n")
        fid.write("N, mm, KJ, C\n")
        fid.write("*SFUNCTION    ; Spectrum Function\n")
        fid.write("   FUNC=E1DRS, 1, 0, 1, 9806, %.2f, , 1.000000\n" % spa_e1.damper_ratio)
        fid.write(spa_e1.parameters)
        i = 1
        for t, s in zip(spa_e1.ts, spa_e1.ss):
            fid.write("    %8.5f, %12.9f" % (t, s))
            if i % 2 == 0:
                fid.write("\n")
            else:
                fid.write(",    ")
            i += 1
        fid.write("   FUNC=E2DRS, 1, 0, 1, 9806, %.2f, , 1.000000\n" % spa_e2.damper_ratio)
        fid.write(spa_e2.parameters)
        i = 1
        for t, s in zip(spa_e2.ts, spa_e2.ss):
            fid.write("    %8.5f, %12.9f" % (t, s))
            if i % 2 == 0:
                fid.write("\n")
            else:
                fid.write(",    ")
            i += 1

        fid.write("*SPLDCASE \n")
        for E in ['E1', 'E2']:
            for direction in ['X', "Y"]:
                fid.write("   NAME=%s%s, XY, 0, 1, 1, NO, NO, LOG,\n" % (E, direction))
                fid.write("        CQC, NO, 0, YES\n")
                fid.write("        %sDRS\n        " % E)
                for j in range(240):
                    fid.write("YES, 1")
                    if j != 239:
                        fid.write(",")
                    else:
                        fid.write("\n")
        fid.write("*EIGEN-CTRL    ; Eigenvalue Analysis Control\n")
        fid.write("   RITZ, NO, 0\n")
        fid.write("   GROUND, ACCX, 80, GROUND, ACCY, 80, GROUND, ACCZ, 80\n")

    def clear(self):
        if self.init:
            shutil.rmtree(self.cwd)
        pass
